from collections.abc import Hashable
from graphlib import TopologicalSorter
from pathlib import Path
from typing import overload

from cognite.client.exceptions import CogniteAPIError, CogniteDuplicatedError
from cognite.client.utils._identifier import T_ID
from rich import print
from rich.markup import escape
from rich.panel import Panel

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands._base import ToolkitCommand
from cognite_toolkit._cdf_tk.commands.clean import CleanCommand
from cognite_toolkit._cdf_tk.constants import (
    _RUNNING_IN_BROWSER,
    BUILD_ENVIRONMENT_FILE,
    HINT_LEAD_TEXT,
)
from cognite_toolkit._cdf_tk.cruds import (
    DataCRUD,
    Loader,
    RawDatabaseCRUD,
    ResourceContainerCRUD,
    ResourceCRUD,
    ResourceWorker,
)
from cognite_toolkit._cdf_tk.cruds._worker import CategorizedResources
from cognite_toolkit._cdf_tk.data_classes import (
    BuildEnvironment,
    DatapointDeployResult,
    DeployResult,
    DeployResults,
    ResourceContainerDeployResult,
    ResourceDeployResult,
    UploadDeployResult,
)
from cognite_toolkit._cdf_tk.data_classes._module_directories import ReadModule
from cognite_toolkit._cdf_tk.exceptions import (
    ResourceCreationError,
    ResourceUpdateError,
    ToolkitDeployResourceError,
    ToolkitFileNotFoundError,
    ToolkitNotADirectoryError,
)
from cognite_toolkit._cdf_tk.protocols import (
    T_ResourceRequest,
    T_ResourceRequestList,
    T_ResourceResponse,
    T_ResourceResponseList,
)
from cognite_toolkit._cdf_tk.tk_warnings import EnvironmentVariableMissingWarning
from cognite_toolkit._cdf_tk.tk_warnings.base import WarningList, catch_warnings
from cognite_toolkit._cdf_tk.tk_warnings.other import (
    LowSeverityWarning,
    ToolkitDependenciesIncludedWarning,
)
from cognite_toolkit._cdf_tk.utils import humanize_collection, read_yaml_file
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables

from ._utils import _print_ids_or_length


class DeployCommand(ToolkitCommand):
    def __init__(self, print_warning: bool = True, skip_tracking: bool = False, silent: bool = False) -> None:
        super().__init__(print_warning, skip_tracking, silent)
        self._clean_command = CleanCommand(print_warning, skip_tracking=True)

    def deploy_build_directory(
        self,
        env_vars: EnvironmentVariables,
        build_dir: Path,
        build_env_name: str | None,
        dry_run: bool,
        drop: bool,
        drop_data: bool,
        force_update: bool,
        include: list[str] | None,
        verbose: bool,
    ) -> None:
        include = self._clean_command.validate_include(include)

        build = self._load_build(build_dir, build_env_name)

        client = env_vars.get_client(build.is_strict_validation)

        self._additional_tracking_info.project = client.config.project
        self._additional_tracking_info.cluster = client.config.cdf_cluster

        if not dry_run:
            for module in build.read_modules:
                if module.module_id:
                    self._additional_tracking_info.module_ids.add(module.module_id)
                if module.package_id:
                    self._additional_tracking_info.package_ids.add(module.package_id)

        selected_loaders = self._clean_command.get_selected_loaders(build_dir, build.read_resource_folders, include)
        ordered_loaders = self._order_loaders(selected_loaders, build_dir)
        self._start_message(build_dir, dry_run, env_vars)
        results = DeployResults([], "deploy", dry_run=dry_run)
        if drop or drop_data:
            self.clean_all_resources(
                client, results, build, build_dir, drop, drop_data, dry_run, env_vars, ordered_loaders, verbose
            )

        self.deploy_all_resources(
            client=client,
            build_dir=build_dir,
            env_vars=env_vars,
            results=results,
            dry_run=dry_run,
            has_dropped=drop,
            has_dropped_data=drop_data,
            force_update=force_update,
            ordered_loaders=ordered_loaders,
            build=build,
            verbose=verbose,
        )

        if results.has_counts:
            print(results.counts_table())
        if results.has_uploads:
            print(results.uploads_table())

    def _load_build(self, build_dir: Path, build_env_name: str | None) -> BuildEnvironment:
        if not build_dir.is_dir():
            raise ToolkitNotADirectoryError(
                "The build directory does not exists. Did you forget to run `cdf build` first?"
            )
        build_environment_file_path = build_dir / BUILD_ENVIRONMENT_FILE
        if not build_environment_file_path.is_file():
            raise ToolkitFileNotFoundError(
                f"Could not find build environment file '{BUILD_ENVIRONMENT_FILE}' in '{build_dir}'. "
                "Did you forget to run `cdf build` first?"
            )
        build = BuildEnvironment.load(read_yaml_file(build_environment_file_path), build_env_name, "deploy")
        build.set_environment_variables()
        errors = build.check_source_files_changed()
        for error in errors:
            self.warn(error)
        if errors:
            raise ToolkitDeployResourceError(
                "One or more source files have been modified since the last build. Please rebuild the project."
            )
        return build

    def _order_loaders(
        self, selected_loaders: dict[type[Loader], frozenset[type[Loader]]], build_dir: Path
    ) -> list[type[Loader]]:
        """The loaders must be topological sorted to ensure that dependencies are deployed in the correct order."""
        ordered_loaders: list[type[Loader]] = []
        should_include: list[type[Loader]] = []
        # The topological sort can include loaders that are not selected, so we need to check for that.
        for loader_cls in TopologicalSorter(selected_loaders).static_order():
            if loader_cls in selected_loaders:
                ordered_loaders.append(loader_cls)
            elif (build_dir / loader_cls.folder_name).is_dir():
                should_include.append(loader_cls)
            # Otherwise, it is not in the build directory and not selected, so we skip it.
            # There should be a warning in the build step if it is missing.
        if should_include:
            self.warn(ToolkitDependenciesIncludedWarning(list({item.folder_name for item in should_include})))
        return ordered_loaders

    @staticmethod
    def _start_message(build_dir: Path, dry_run: bool, env_vars: EnvironmentVariables) -> None:
        environment_vars = ""
        if not _RUNNING_IN_BROWSER:
            environment_vars = f"\n\nConnected to {env_vars.as_string()}"
        verb = "Checking" if dry_run else "Deploying"
        print(
            Panel(
                f"[bold]{verb}[/]resource files from {build_dir} directory.{environment_vars}",
                expand=False,
            )
        )

    def clean_all_resources(
        self,
        client: ToolkitClient,
        results: DeployResults,
        build: BuildEnvironment,
        build_dir: Path,
        drop: bool,
        drop_data: bool,
        dry_run: bool,
        env_vars: EnvironmentVariables,
        ordered_loaders: list[type[Loader]],
        verbose: bool,
    ) -> None:
        # Drop has to be done in the reverse order of deploy.
        if drop and drop_data:
            print(Panel("[bold] Cleaning resources as --drop and --drop-data are passed[/]"))
        elif drop:
            print(Panel("[bold] Cleaning resources as --drop is passed[/]"))
        elif drop_data:
            print(Panel("[bold] Cleaning resources as --drop-data is passed[/]"))
        else:
            return None

        for loader_cls in reversed(ordered_loaders):
            if not issubclass(loader_cls, ResourceCRUD):
                continue
            loader: ResourceCRUD = loader_cls.create_loader(client, build_dir)
            result = self._clean_command.clean_resources(
                loader,
                env_vars=env_vars,
                read_modules=build.read_modules,
                drop=drop,
                dry_run=dry_run,
                drop_data=drop_data,
                verbose=verbose,
            )
            if result:
                results[result.name] = result
        print("[bold]...cleaning complete![/]")
        return None

    @overload
    def deploy_all_resources(
        self,
        client: ToolkitClient,
        build_dir: Path,
        env_vars: EnvironmentVariables,
        results: None = None,
        dry_run: bool = False,
        has_dropped: bool = False,
        has_dropped_data: bool = False,
        force_update: bool = False,
        ordered_loaders: list[type[Loader]] | None = None,
        build: BuildEnvironment | None = None,
        verbose: bool = False,
    ) -> DeployResults: ...

    @overload
    def deploy_all_resources(
        self,
        client: ToolkitClient,
        build_dir: Path,
        env_vars: EnvironmentVariables,
        results: DeployResults,
        dry_run: bool = False,
        has_dropped: bool = False,
        has_dropped_data: bool = False,
        force_update: bool = False,
        ordered_loaders: list[type[Loader]] | None = None,
        build: BuildEnvironment | None = None,
        verbose: bool = False,
    ) -> None: ...

    def deploy_all_resources(
        self,
        client: ToolkitClient,
        build_dir: Path,
        env_vars: EnvironmentVariables,
        results: DeployResults | None = None,
        dry_run: bool = False,
        has_dropped: bool = False,
        has_dropped_data: bool = False,
        force_update: bool = False,
        ordered_loaders: list[type[Loader]] | None = None,
        build: BuildEnvironment | None = None,
        verbose: bool = False,
    ) -> DeployResults | None:
        """Deploy all resources in the build directory.

        Args:
            client: The Cognite client to use for deployment.
            build_dir: The directory containing the built resource files.
            env_vars: The environment variables to use for variable replacement in the resource files.
            results: An optional DeployResults object to store the results of the deployment. If None, a new object will be created and returned.
            dry_run: If True, performs a dry run without actually deploying the resources.
            has_dropped: If True, indicates that resources have been dropped prior to deployment.
            has_dropped_data: If True, indicates that data resources have been dropped prior to deployment.
            force_update: If True, forces an update of resources even if they are unchanged.
            ordered_loaders: An optional list of loader classes to use for deployment. If None, all loaders will be used in the order determined by dependencies.
            build: An optional BuildEnvironment object containing information about the build. If None, data uploads will be skipped.
            verbose: If True, prints detailed information about the deployment process.

        Returns:
            A DeployResults object containing the results of the deployment if results is None, otherwise None.

        """
        if verbose:
            print(Panel("[bold]DEPLOYING resources...[/]"))

        if ordered_loaders is None:
            selected_loaders = self._clean_command.get_selected_loaders(build_dir, set(), None)
            ordered_loaders = self._order_loaders(selected_loaders, build_dir)
        read_modules: list[ReadModule] | None = None
        if build is not None:
            read_modules = build.read_modules
        output_results = DeployResults([], "deploy", dry_run=dry_run) if results is None else results
        for loader_cls in ordered_loaders:
            loader = loader_cls.create_loader(client, build_dir)
            resource_result: DeployResult | None
            if isinstance(loader, ResourceCRUD):
                resource_result = self.deploy_resource_type(
                    loader,
                    env_vars,
                    read_modules,
                    dry_run,
                    has_dropped,
                    has_dropped_data,
                    force_update,
                    verbose,
                )
            elif isinstance(loader, DataCRUD) and build is not None:
                resource_result = self.upload_data(loader, build, dry_run, verbose)
            elif isinstance(loader, DataCRUD) and build is None:
                self.warn(
                    LowSeverityWarning(
                        f"Skipping data upload for {loader.display_name} as it is not supported in this context."
                    )
                )
                continue
            else:
                raise ValueError(f"Unsupported loader type {type(loader)}.")

            if resource_result:
                output_results[resource_result.name] = resource_result
            if verbose:
                print("")  # Extra newline
        if results is None:
            return output_results
        return None

    def deploy_resource_type(
        self,
        loader: ResourceCRUD[
            T_ID, T_ResourceRequest, T_ResourceResponse, T_ResourceRequestList, T_ResourceResponseList
        ],
        env_vars: EnvironmentVariables,
        read_modules: list[ReadModule] | None = None,
        dry_run: bool = False,
        has_done_drop: bool = False,
        has_dropped_data: bool = False,
        force_update: bool = False,
        verbose: bool = False,
    ) -> ResourceDeployResult | None:
        worker = ResourceWorker(loader, "deploy")
        files = worker.load_files(read_modules=read_modules)
        if not files:
            return None

        with catch_warnings(EnvironmentVariableMissingWarning) as env_var_warnings:
            resources = worker.prepare_resources(
                files,
                environment_variables=env_vars.dump(include_os=True),
                is_dry_run=dry_run,
                force_update=force_update,
                verbose=verbose,
            )
        if env_var_warnings:
            print(str(env_var_warnings))
            self.warning_list.extend(env_var_warnings)

        # We are not counting to_delete as these are captured by to_create.
        # (to_delete is used for resources that does not support update and instead needs to be deleted and recreated)
        nr_of_items = len(resources.to_create) + len(resources.to_update) + len(resources.unchanged)
        if nr_of_items == 0:
            return ResourceDeployResult(name=loader.display_name)

        prefix = "Would deploy" if dry_run else "Deploying"
        print(f"[bold]{prefix} {nr_of_items} {loader.display_name} to CDF...[/]")

        # Moved here to avoid printing before the above message.
        if not isinstance(loader, RawDatabaseCRUD):
            for duplicate in worker.duplicates:
                self.warn(LowSeverityWarning(f"Skipping duplicate {loader.display_name} {duplicate}."))

        if dry_run:
            result = self.dry_run_deploy(
                resources,
                loader,
                has_done_drop,
                has_dropped_data,
            )
        else:
            result = self.actual_deploy(resources, loader, env_var_warnings)

        if verbose:
            self._verbose_print(resources, loader, dry_run)

        if isinstance(loader, ResourceContainerCRUD):
            return ResourceContainerDeployResult.from_resource_deploy_result(
                result,
                item_name=loader.item_name,
            )
        return result

    def actual_deploy(
        self,
        resources: CategorizedResources[T_ID, T_ResourceResponseList],
        loader: ResourceCRUD[
            T_ID, T_ResourceRequest, T_ResourceResponse, T_ResourceRequestList, T_ResourceResponseList
        ],
        env_var_warnings: WarningList | None = None,
    ) -> ResourceDeployResult:
        environment_variable_warning_by_id = {
            identifier: warning
            for warning in env_var_warnings or []
            if isinstance(warning, EnvironmentVariableMissingWarning)
            for identifier in warning.identifiers or []
        }
        nr_of_unchanged = len(resources.unchanged)
        nr_of_deleted, nr_of_created, nr_of_changed = 0, 0, 0
        if resources.to_delete:
            deleted = loader.delete(resources.to_delete)
            nr_of_deleted += deleted
        if resources.to_create:
            created = self._create_resources(resources.to_create, loader, environment_variable_warning_by_id)
            nr_of_created += created
        if resources.to_update:
            updated = self._update_resources(resources.to_update, loader, environment_variable_warning_by_id)
            nr_of_changed += updated
        return ResourceDeployResult(
            name=loader.display_name,
            created=nr_of_created,
            deleted=nr_of_deleted,
            changed=nr_of_changed,
            unchanged=nr_of_unchanged,
            total=nr_of_created + nr_of_deleted + nr_of_changed + nr_of_unchanged,
        )

    @staticmethod
    def dry_run_deploy(
        resources: CategorizedResources[T_ID, T_ResourceResponseList],
        loader: ResourceCRUD[
            T_ID, T_ResourceRequest, T_ResourceResponse, T_ResourceRequestList, T_ResourceResponseList
        ],
        has_done_drop: bool,
        has_dropped_data: bool,
    ) -> ResourceDeployResult:
        if (
            loader.support_drop
            and has_done_drop
            and (not isinstance(loader, ResourceContainerCRUD) or has_dropped_data)
        ):
            # Means the resources will be deleted and not left unchanged or changed
            for item in resources.unchanged:
                # We cannot use extents as LoadableNodes cannot be extended.
                resources.to_create.append(item)
            for item in resources.to_update:
                resources.to_create.append(item)
            resources.unchanged.clear()
            resources.to_update.clear()
        return ResourceDeployResult(
            name=loader.display_name,
            created=len(resources.to_create),
            deleted=len(resources.to_delete),
            changed=len(resources.to_update),
            unchanged=len(resources.unchanged),
            total=len(resources.to_create)
            + len(resources.to_delete)
            + len(resources.to_update)
            + len(resources.unchanged),
        )

    @staticmethod
    def _verbose_print(
        resources: CategorizedResources[T_ID, T_ResourceResponseList],
        loader: ResourceCRUD,
        dry_run: bool,
    ) -> None:
        print_outs = []
        prefix = "Would have " if dry_run else ""
        if resources.to_create:
            print_outs.append(f"{prefix}Created {_print_ids_or_length(loader.get_ids(resources.to_create), limit=20)}")
        if resources.to_update:
            print_outs.append(f"{prefix}Updated {_print_ids_or_length(loader.get_ids(resources.to_update), limit=20)}")
        if resources.unchanged:
            print_outs.append(
                f"{'Untouched' if dry_run else 'Unchanged'} {_print_ids_or_length(loader.get_ids(resources.unchanged), limit=5)}"
            )
        prefix_message = f" {loader.display_name}: "
        if len(print_outs) == 1:
            print(f"{prefix_message}{print_outs[0]}")
        elif len(print_outs) == 2:
            print(f"{prefix_message}{print_outs[0]} and {print_outs[1]}")
        else:
            print(f"{prefix_message}{', '.join(print_outs[:-1])} and {print_outs[-1]}")

    def _create_resources(
        self,
        resources: T_ResourceResponseList,
        loader: ResourceCRUD,
        environment_variable_warning_by_id: dict[Hashable, EnvironmentVariableMissingWarning],
    ) -> int:
        try:
            created = loader.create(resources)
        except CogniteAPIError as e:
            message = f"Failed to create resource(s). Error: {escape(str(e))!s}."
            if hint := self._environment_variable_hint(loader.get_ids(resources), environment_variable_warning_by_id):
                message += hint
            raise ResourceCreationError(message) from e
        except CogniteDuplicatedError as e:
            self.warn(
                LowSeverityWarning(
                    f"{len(e.duplicated)} out of {len(resources)} resource(s) already exist(s). {len(e.successful or [])} resource(s) created."
                )
            )
        else:
            return len(created) if created is not None else 0
        return 0

    def _update_resources(
        self,
        resources: T_ResourceResponseList,
        loader: ResourceCRUD,
        environment_variable_warning_by_id: dict[Hashable, EnvironmentVariableMissingWarning],
    ) -> int:
        try:
            updated = loader.update(resources)
        except CogniteAPIError as e:
            message = f"Failed to update resource(s). Error: {escape(str(e))}."
            if hint := self._environment_variable_hint(loader.get_ids(resources), environment_variable_warning_by_id):
                message += hint
            raise ResourceUpdateError(message) from e

        return len(updated)

    @staticmethod
    def _environment_variable_hint(
        identifiers: list[Hashable],
        environment_variable_warning_by_id: dict[Hashable, EnvironmentVariableMissingWarning],
    ) -> str:
        if not environment_variable_warning_by_id:
            return ""
        missing_variables: set[str] = set()
        for identifier in identifiers:
            if warning := environment_variable_warning_by_id.get(identifier):
                missing_variables.update(warning.variables)
        if missing_variables:
            variables_str = humanize_collection(missing_variables)
            suffix = "s" if len(missing_variables) > 1 else ""
            return f"\n  {HINT_LEAD_TEXT}This is likely due to missing environment variable{suffix}: {variables_str}"
        return ""

    @staticmethod
    def upload_data(
        loader: DataCRUD,
        state: BuildEnvironment,
        dry_run: bool = False,
        verbose: bool = False,
    ) -> UploadDeployResult:
        prefix = "Would upload" if dry_run else "Uploading"
        print(f"[bold]{prefix} {loader.display_name} files to CDF...[/]")

        datapoints = 0
        file_counts = 0
        for message, file_datapoints in loader.upload(state, dry_run):
            if verbose:
                print(message)
            datapoints += file_datapoints
            file_counts += 1

        if datapoints != 0:
            return DatapointDeployResult(
                loader.display_name, points=datapoints, uploaded=file_counts, item_name=loader.item_name
            )
        else:
            return UploadDeployResult(loader.display_name, uploaded=file_counts, item_name=loader.item_name)



from pathlib import Path

import pytest
from pydantic import BaseModel, Field, field_validator

from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError
from cognite_toolkit._cdf_tk.storageio import InstanceIdCSVList, InstanceIdRow, ModelList


class MyRow(BaseModel):
    id: int
    name: str
    display_name: str | None = Field(None, alias="displayName")

    @field_validator("display_name", mode="before")
    @classmethod
    def validate_display_name(cls, v: str | None) -> str | None:
        if isinstance(v, str) and v.strip() == "":
            return None
        return v


class MyList(ModelList[MyRow]):
    @classmethod
    def _get_base_model_cls(cls) -> type[MyRow]:
        return MyRow


@pytest.fixture()
def a_list() -> MyList:
    return MyList(
        [
            MyRow(id=1, name="Alice", displayName="Aly"),
            MyRow(id=2, name="Bob"),
            MyRow(id=3, name="Charlie"),
        ]
    )


class TestModelList:
    def test_read_csv(self, tmp_path: Path) -> None:
        csv_content = "id,name,displayName\n1,Alice,Aly\n2,Bob,\ninvalid_row\n3,Charlie,\n"
        csv_file = tmp_path / "test.csv"
        csv_file.write_text(csv_content)

        my_list = MyList.read_csv_file(csv_file)

        assert len(my_list) == 3
        assert my_list[0] == MyRow(id=1, name="Alice", displayName="Aly")
        assert my_list[1] == MyRow(id=2, name="Bob")
        assert my_list[2] == MyRow(id=3, name="Charlie")
        assert len(my_list.invalid_rows) == 1
        assert 3 in my_list.invalid_rows

    def test_raise_missing_column(self, tmp_path: Path) -> None:
        csv_content = "id\n1\n2\n3\n"
        csv_file = tmp_path / "test_missing_column.csv"
        csv_file.write_text(csv_content)

        with pytest.raises(ToolkitValueError, match="Missing required columns: name"):
            MyList.read_csv_file(csv_file)

    def test_iterate(self, a_list: MyList) -> None:
        names = [row.name for row in a_list]
        assert names == ["Alice", "Bob", "Charlie"]

    def test_get_item(self, a_list: MyList) -> None:
        assert a_list[0] == MyRow(id=1, name="Alice", displayName="Aly")
        assert a_list[1:3] == MyList([MyRow(id=2, name="Bob"), MyRow(id=3, name="Charlie")])


class TestInstanceCSVList:
    def test_read_csv(self, tmp_path: Path) -> None:
        csv_content = "space,externalId,instanceType\nmySpace,id1,node\nmySpace,id2,edge\n"
        csv_file = tmp_path / "instances.csv"
        csv_file.write_text(csv_content)

        instance_list = InstanceIdCSVList.read_csv_file(csv_file)

        assert len(instance_list) == 2
        assert instance_list[0] == InstanceIdRow(externalId="id1", space="mySpace")
        assert instance_list[1] == InstanceIdRow(externalId="id2", space="mySpace", instanceType="edge")
        assert len(instance_list.invalid_rows) == 0



from pathlib import Path
from unittest.mock import MagicMock

from cognite.client.data_classes.data_modeling import Container, ContainerProperty, Text

from cognite_toolkit._cdf_tk.cruds import ContainerCRUD, ResourceWorker
from tests.test_unit.approval_client import ApprovalToolkitClient


class TestContainerLoader:
    def test_unchanged_used_for_not_set(self, toolkit_client_approval: ApprovalToolkitClient) -> None:
        loader = ContainerCRUD.create_loader(toolkit_client_approval.mock_client)
        raw_file = """space: sp_enterprise_process_industry_full
externalId: Toolkit360Image
properties:
  UUID:
    type:
      list: false
      collation: ucs_basic
      type: text
    immutable: false
    nullable: true
    autoIncrement: false
constraints: {}
indexes: {}
"""
        file = MagicMock(spec=Path)
        file.read_text.return_value = raw_file
        cdf_container = Container(
            space="sp_enterprise_process_industry_full",
            external_id="Toolkit360Image",
            last_updated_time=1739469813633,
            created_time=1739469813633,
            description=None,
            name=None,
            used_for="node",
            is_global=False,
            properties={"UUID": ContainerProperty(type=Text())},
            indexes={},
            constraints={},
        )

        toolkit_client_approval.append(Container, [cdf_container])

        worker = ResourceWorker(loader, "deploy")
        resources = worker.prepare_resources([file])
        assert {
            "create": len(resources.to_create),
            "change": len(resources.to_update),
            "delete": len(resources.to_delete),
            "unchanged": len(resources.unchanged),
        } == {"create": 0, "change": 0, "delete": 0, "unchanged": 1}



import importlib.util
from typing import TYPE_CHECKING

from cognite_toolkit._cdf_tk.exceptions import ToolkitMissingDependencyError

if TYPE_CHECKING:
    from sqlparse.sql import Identifier
    from sqlparse.tokens import Token
from dataclasses import dataclass


@dataclass(frozen=True)
class SQLTable:
    schema: str
    name: str

    def __str__(self) -> str:
        """Return the table name in the format 'schema.table'."""
        return f"{self.schema}.{self.name}"


class SQLParser:
    def __init__(self, query: str, operation: str) -> None:
        self._verify_dependencies(operation)
        self.query = query
        self._seen_sources: set[SQLTable] = set()
        self._sources: list[SQLTable] = []
        self._destination_columns: list[str] = []
        self._is_parsed = False

    @staticmethod
    def _verify_dependencies(operation: str) -> None:
        if importlib.util.find_spec("sqlparse") is None:
            raise ToolkitMissingDependencyError(
                f"{operation} requires sqlparse. Install with 'pip install \"cognite-toolkit[sql]\"'"
            )

    @property
    def sources(self) -> list[SQLTable]:
        """Returns a list of sources (tables) found in the SQL query."""
        if not self._is_parsed:
            self.parse()
        return self._sources

    @property
    def destination_columns(self) -> list[str]:
        """Returns a list of destination columns found in the SQL query."""
        if not self._is_parsed:
            self.parse()
        return self._destination_columns

    def is_using_data_set(
        self, data_set_ids: list[int] | None = None, data_set_external_ids: list[str] | None = None
    ) -> bool:
        """Check if the SQL query uses any of the provided data set IDs or external IDs."""
        for data_set_id in data_set_ids or []:
            if f"{data_set_id} " in self.query:
                return True
        for data_set_external_id in data_set_external_ids or []:
            if f'dataset_id("{data_set_external_id}")' in self.query:
                return True
            elif f"dataset_id('{data_set_external_id}')" in self.query:
                return True
        return False

    def parse(self) -> None:
        """Parse the SQL query and extract table names."""
        import sqlparse

        parsed = sqlparse.parse(self.query)
        if not parsed:
            self._is_parsed = True
            return

        for statement in parsed:
            self._find_tables(statement.tokens)
            self._find_destination_columns(statement.tokens)
        self._is_parsed = True
        return

    def _find_destination_columns(self, tokens: "list[Token]") -> None:
        from sqlparse.sql import Comment, Identifier, IdentifierList
        from sqlparse.tokens import DML, Wildcard

        content_tokens = [
            token
            for token in tokens
            if not token.is_whitespace and not token.is_newline and not isinstance(token, Comment)
        ]
        is_next: bool = False
        for token in content_tokens:
            if is_next and isinstance(token, IdentifierList):
                self._add_destination_columns(*token.get_identifiers())
                break
            elif is_next and isinstance(token, Identifier):
                self._add_destination_columns(token)
            elif is_next and token.ttype is Wildcard and token.normalized == "*":
                self._destination_columns.append("*")
                break
            elif is_next and token.is_keyword:
                break
            if token.ttype is DML and token.normalized == "SELECT":
                is_next = True

    def _add_destination_columns(self, *identifiers: "Identifier") -> None:
        from sqlparse.sql import Comment, Identifier

        for identifier in identifiers:
            # The identifier is either a `column_name` or `some expression AS column_name`.
            # Thus, we simply take the last token in the identifier.
            if isinstance(identifier, Identifier) and identifier.tokens:
                content_tokens = [
                    token
                    for token in identifier.tokens
                    if not token.is_whitespace and not token.is_newline and not isinstance(token, Comment)
                ]
                if content_tokens:
                    self._destination_columns.append(content_tokens[-1].value)

    def _find_tables(self, tokens: "list[Token]") -> None:
        from sqlparse.sql import Identifier, IdentifierList, TokenList
        from sqlparse.tokens import Keyword

        content_tokens = [token for token in tokens if not token.is_whitespace and not token.is_newline]

        is_next_source: bool = False
        for token in content_tokens:
            if is_next_source and isinstance(token, Identifier):
                if not self._add_to_source(token):
                    # This could be a nested expression like '(SELECT ... FROM ...) Identifier'.
                    self._find_tables(token.tokens)
                is_next_source = False
            elif is_next_source and isinstance(token, IdentifierList):
                for identifier in token.get_identifiers():
                    if not self._add_to_source(identifier):
                        self._find_tables(identifier.tokens)
            elif token.ttype is Keyword and (token.normalized == "FROM" or token.normalized.endswith("JOIN")):
                is_next_source = True
            elif isinstance(token, TokenList):
                # If the token is a TokenList, we need to check its tokens recursively.
                self._find_tables(token.tokens)

    def _add_to_source(self, source: "Identifier") -> bool:
        """Add a source to the list if it hasn't been seen before."""
        from sqlparse.tokens import Name

        names = [token for token in source.tokens if token.ttype is Name]

        if len(names) != 2:
            # We are expecting a schema and a table name
            return False
        table = SQLTable(schema=self._clean_name(names[0].value), name=self._clean_name(names[1].value))
        if table not in self._seen_sources:
            self._seen_sources.add(table)
            self._sources.append(table)
        return True

    @staticmethod
    def _clean_name(name: str) -> str:
        """Clean the name by removing quotes and whitespace."""
        return name.strip().removeprefix("`").removesuffix("`").strip()



import uuid
from abc import ABC, abstractmethod
from collections.abc import Callable, Hashable, Iterable, Sequence
from dataclasses import dataclass
from functools import partial
from typing import Literal, cast

import questionary
from cognite.client.data_classes import DataSetUpdate
from cognite.client.data_classes._base import CogniteResourceList
from cognite.client.data_classes.data_modeling import (
    EdgeList,
    NodeId,
    NodeList,
)
from cognite.client.data_classes.data_modeling.statistics import SpaceStatistics
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils._identifier import InstanceId
from rich import print
from rich.console import Console
from rich.panel import Panel

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.cruds import (
    AssetCRUD,
    ContainerCRUD,
    DataModelCRUD,
    EdgeCRUD,
    EventCRUD,
    ExtractionPipelineCRUD,
    FileMetadataCRUD,
    LabelCRUD,
    NodeCRUD,
    RelationshipCRUD,
    ResourceCRUD,
    SequenceCRUD,
    SpaceCRUD,
    ThreeDModelCRUD,
    TimeSeriesCRUD,
    TransformationCRUD,
    ViewCRUD,
    WorkflowCRUD,
)
from cognite_toolkit._cdf_tk.data_classes import DeployResults, ResourceDeployResult
from cognite_toolkit._cdf_tk.exceptions import (
    AuthorizationError,
    ToolkitMissingResourceError,
)
from cognite_toolkit._cdf_tk.storageio import InstanceIO
from cognite_toolkit._cdf_tk.storageio.selectors import InstanceSelector
from cognite_toolkit._cdf_tk.tk_warnings import (
    HighSeverityWarning,
    LimitedAccessWarning,
)
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils.aggregators import (
    AssetAggregator,
    EventAggregator,
    FileAggregator,
    LabelCountAggregator,
    RelationshipAggregator,
    SequenceAggregator,
    TimeSeriesAggregator,
)
from cognite_toolkit._cdf_tk.utils.http_client import (
    FailedRequestItems,
    FailedResponseItems,
    HTTPClient,
    ItemsRequest,
    SuccessResponseItems,
)
from cognite_toolkit._cdf_tk.utils.producer_worker import ProducerWorkerExecutor
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal
from cognite_toolkit._cdf_tk.utils.validate_access import ValidateAccess

from ._base import ToolkitCommand


@dataclass
class DeleteResults:
    deleted: int = 0
    failed: int = 0


@dataclass
class DeleteItem:
    item: JsonVal
    as_id_fun: Callable[[JsonVal], Hashable]

    def dump(self) -> JsonVal:
        return self.item

    def as_id(self) -> Hashable:
        return self.as_id_fun(self.item)


@dataclass
class ToDelete(ABC):
    crud: ResourceCRUD
    total: int
    delete_url: str

    @property
    def display_name(self) -> str:
        return self.crud.display_name

    @abstractmethod
    def get_process_function(
        self, client: ToolkitClient, console: Console, verbose: bool, process_results: ResourceDeployResult
    ) -> Callable[[CogniteResourceList], list[JsonVal]]:
        raise NotImplementedError()

    def get_extra_fields(self) -> dict[str, JsonVal]:
        return {}


@dataclass
class DataModelingToDelete(ToDelete):
    def get_process_function(
        self, client: ToolkitClient, console: Console, verbose: bool, process_results: ResourceDeployResult
    ) -> Callable[[CogniteResourceList], list[JsonVal]]:
        def as_id(chunk: CogniteResourceList) -> list[JsonVal]:
            return [item.as_id().dump(include_type=False) for item in chunk]

        return as_id


@dataclass
class EdgeToDelete(ToDelete):
    def get_process_function(
        self, client: ToolkitClient, console: Console, verbose: bool, process_results: ResourceDeployResult
    ) -> Callable[[CogniteResourceList], list[JsonVal]]:
        def as_id(chunk: CogniteResourceList) -> list[JsonVal]:
            return [
                {"space": item.space, "externalId": item.external_id, "instanceType": "edge"}
                for item in cast(EdgeList, chunk)
            ]

        return as_id


@dataclass
class NodesToDelete(ToDelete):
    delete_datapoints: bool
    delete_file_content: bool

    def get_process_function(
        self, client: ToolkitClient, console: Console, verbose: bool, process_results: ResourceDeployResult
    ) -> Callable[[CogniteResourceList], list[JsonVal]]:
        def check_for_data(chunk: CogniteResourceList) -> list[JsonVal]:
            node_ids = cast(NodeList, chunk).as_ids()
            found_ids: set[InstanceId] = set()
            if not self.delete_datapoints:
                timeseries = client.time_series.retrieve_multiple(instance_ids=node_ids, ignore_unknown_ids=True)
                found_ids |= {ts.instance_id for ts in timeseries if ts.instance_id is not None}
            if not self.delete_file_content:
                files = client.files.retrieve_multiple(instance_ids=node_ids, ignore_unknown_ids=True)
                found_ids |= {f.instance_id for f in files if f.instance_id is not None}
            if found_ids and verbose:
                console.print(f"Skipping {found_ids} nodes as they have datapoints or file content")
            process_results.unchanged += len(found_ids)
            result: list[JsonVal] = []
            for node_id in (n for n in node_ids if n not in found_ids):
                dumped = node_id.dump(include_instance_type=True)
                # The delete endpoint expects "instanceType" instead of "type"
                dumped["instanceType"] = dumped.pop("type")
                # MyPy think complains about invariant here, even though dict[str, str] is a type of JsonVal
                result.append(dumped)  # type: ignore[arg-type]
            return result

        return check_for_data


@dataclass
class IdResourceToDelete(ToDelete):
    def get_process_function(
        self, client: ToolkitClient, console: Console, verbose: bool, process_results: ResourceDeployResult
    ) -> Callable[[CogniteResourceList], list[JsonVal]]:
        def as_id(chunk: CogniteResourceList) -> list[JsonVal]:
            return [{"id": item.id} for item in chunk]

        return as_id


@dataclass
class ExternalIdToDelete(ToDelete):
    def get_process_function(
        self, client: ToolkitClient, console: Console, verbose: bool, process_results: ResourceDeployResult
    ) -> Callable[[CogniteResourceList], list[JsonVal]]:
        def as_external_id(chunk: CogniteResourceList) -> list[JsonVal]:
            return [{"externalId": item.external_id} for item in chunk]

        return as_external_id


@dataclass
class AssetToDelete(IdResourceToDelete):
    recursive: bool

    def get_extra_fields(self) -> dict[str, JsonVal]:
        return {"recursive": self.recursive}


class PurgeCommand(ToolkitCommand):
    BATCH_SIZE_DM = 1000

    def space(
        self,
        client: ToolkitClient,
        selected_space: str,
        include_space: bool = False,
        delete_datapoints: bool = False,
        delete_file_content: bool = False,
        dry_run: bool = False,
        auto_yes: bool = False,
        verbose: bool = False,
    ) -> DeployResults:
        # Warning Messages
        if not dry_run:
            self._print_panel("space", selected_space)

        if not dry_run and not auto_yes:
            confirm = questionary.confirm(
                f"Are you really sure you want to purge the {selected_space!r} space?", default=False
            ).ask()
            if not confirm:
                return DeployResults([], "purge", dry_run=dry_run)

        stats = client.data_modeling.statistics.spaces.retrieve(selected_space)
        if stats is None:
            raise ToolkitMissingResourceError(f"Space {selected_space!r} does not exist")

        # ValidateAuth
        validator = ValidateAccess(client, "purge")
        if include_space or (stats.containers + stats.views + stats.data_models) > 0:
            # We check for write even in dry-run mode. This is because dry-run is expected to fail
            # if the user cannot perform the purge.
            validator.data_model(["read", "write"], spaces={selected_space})
        if (stats.nodes + stats.edges) > 0:
            validator.instances(["read", "write"], spaces={selected_space})

        to_delete = self._create_to_delete_list_purge_space(client, delete_datapoints, delete_file_content, stats)
        if dry_run:
            results = DeployResults([], "purge", dry_run=True)
            for item in to_delete:
                results[item.display_name] = ResourceDeployResult(item.display_name, deleted=item.total)
            if include_space:
                space_loader = SpaceCRUD.create_loader(client)
                results[space_loader.display_name] = ResourceDeployResult(space_loader.display_name, deleted=1)
        else:
            results = self._delete_resources(to_delete, client, verbose, selected_space, None)
            if include_space:
                self._delete_space(client, selected_space, results)
        print(results.counts_table(exclude_columns={"Created", "Changed", "Total"}))
        return results

    @staticmethod
    def _create_to_delete_list_purge_space(
        client: ToolkitClient, delete_datapoints: bool, delete_file_content: bool, stats: SpaceStatistics
    ) -> list[ToDelete]:
        config = client.config
        to_delete = [
            EdgeToDelete(
                EdgeCRUD.create_loader(client), stats.edges, config.create_api_url("/models/instances/delete")
            ),
            NodesToDelete(
                NodeCRUD.create_loader(client),
                stats.nodes,
                config.create_api_url(
                    "/models/instances/delete",
                ),
                delete_datapoints=delete_datapoints,
                delete_file_content=delete_file_content,
            ),
            DataModelingToDelete(
                DataModelCRUD.create_loader(client),
                stats.data_models,
                config.create_api_url("/models/datamodels/delete"),
            ),
            DataModelingToDelete(
                ViewCRUD.create_loader(client), stats.views, config.create_api_url("/models/views/delete")
            ),
            DataModelingToDelete(
                ContainerCRUD.create_loader(client),
                stats.containers,
                config.create_api_url("/models/containers/delete"),
            ),
        ]
        return to_delete

    def _delete_space(self, client: ToolkitClient, selected_space: str, results: DeployResults) -> None:
        space_loader = SpaceCRUD.create_loader(client)
        try:
            space_loader.delete([selected_space])
            print(f"Space {selected_space} deleted")
        except CogniteAPIError as e:
            self.warn(HighSeverityWarning(f"Failed to delete space {selected_space!r}: {e}"))
        else:
            results[space_loader.display_name] = ResourceDeployResult(space_loader.display_name, deleted=1)

    def _delete_resources(
        self,
        to_delete: list[ToDelete],
        client: ToolkitClient,
        verbose: bool,
        space: str | None,
        data_set_external_id: str | None,
    ) -> DeployResults:
        results = DeployResults([], "purge", dry_run=False)
        console = Console()
        with HTTPClient(client.config, max_retries=10) as delete_client:
            for item in to_delete:
                if item.total == 0:
                    results[item.display_name] = ResourceDeployResult(item.display_name, deleted=0)
                    continue
                # Two results objects since they are updated concurrently
                process_results = ResourceDeployResult(item.display_name)
                write_results = ResourceDeployResult(item.display_name)
                iteration_count: int | None = None
                if item.total > 0:
                    iteration_count = item.total // self.BATCH_SIZE_DM + (
                        1 if item.total % self.BATCH_SIZE_DM > 0 else 0
                    )
                executor = ProducerWorkerExecutor[CogniteResourceList, list[JsonVal]](
                    download_iterable=self._iterate_batch(
                        item.crud, space, data_set_external_id, batch_size=self.BATCH_SIZE_DM
                    ),
                    process=item.get_process_function(client, console, verbose, process_results),
                    write=self._purge_batch(item, item.delete_url, delete_client, write_results),
                    max_queue_size=10,
                    iteration_count=iteration_count,
                    download_description=f"Downloading {item.display_name}",
                    process_description=f"Preparing {item.display_name} for deletion",
                    write_description=f"Deleting {item.display_name}",
                    console=console,
                )
                executor.run()
                write_results += process_results
                results[item.display_name] = write_results
                if executor.error_occurred:
                    if verbose and executor.error_traceback:
                        executor.print_traceback()
                    self.warn(
                        HighSeverityWarning(f"Failed to delete all {item.display_name}. {executor.error_message}")
                    )
        return results

    @staticmethod
    def _iterate_batch(
        crud: ResourceCRUD, selected_space: str | None, data_set_external_id: str | None, batch_size: int
    ) -> Iterable[CogniteResourceList]:
        batch = crud.list_cls([])
        for resource in crud.iterate(space=selected_space, data_set_external_id=data_set_external_id):
            batch.append(resource)
            if len(batch) >= batch_size:
                yield batch
                batch = crud.list_cls([])
        if batch:
            yield batch

    @staticmethod
    def _purge_batch(
        delete_item: ToDelete, delete_url: str, delete_client: HTTPClient, result: ResourceDeployResult
    ) -> Callable[[list[JsonVal]], None]:
        crud = delete_item.crud

        def as_id(item: JsonVal) -> Hashable:
            try:
                return crud.get_id(item)
            except KeyError:
                # Fallback to internal ID
                return crud.get_internal_id(item)

        def process(items: list[JsonVal]) -> None:
            responses = delete_client.request_with_retries(
                ItemsRequest(
                    endpoint_url=delete_url,
                    method="POST",
                    items=[DeleteItem(item=item, as_id_fun=as_id) for item in items],
                    extra_body_fields=delete_item.get_extra_fields(),
                )
            )
            for response in responses:
                if isinstance(response, SuccessResponseItems):
                    result.deleted += len(response.ids)
                else:
                    result.unchanged += len(items)

        return process

    def dataset(
        self,
        client: ToolkitClient,
        selected_data_set_external_id: str,
        archive_dataset: bool = False,
        include_data: bool = True,
        include_configurations: bool = False,
        asset_recursive: bool = False,
        dry_run: bool = False,
        auto_yes: bool = False,
        verbose: bool = False,
    ) -> DeployResults:
        """Purge a dataset and all its content

        Args:
            client: The ToolkitClient to use
            selected_data_set_external_id:  The external ID of the dataset to purge
            archive_dataset:  Whether to archive the dataset itself after the purge
            include_data: Whether to include data (assets, events, time series, files, sequences, 3D models, relationships, labels) in the purge
            include_configurations: Whether to include configurations (workflows, transformations, extraction pipelines) in the purge
            asset_recursive: Whether to recursively delete assets.
            dry_run: Whether to perform a dry run
            auto_yes:  Whether to automatically confirm the purge
            verbose:  Whether to print verbose output

        Returns:
            DeployResults: The results of the purge operation

        """
        # Warning Messages
        if not dry_run:
            self._print_panel("dataSet", selected_data_set_external_id)
        if not dry_run and not auto_yes:
            confirm = questionary.confirm(
                f"Are you really sure you want to purge the {selected_data_set_external_id!r} dataSet?", default=False
            ).ask()
            if not confirm:
                return DeployResults([], "purge", dry_run=dry_run)

        # Validate Auth
        validator = ValidateAccess(client, "purge")
        data_set_id = client.lookup.data_sets.id(selected_data_set_external_id)
        if data_set_id is None:
            raise ToolkitMissingResourceError(f"DataSet {selected_data_set_external_id!r} does not exist")
        action = cast(Sequence[Literal["read", "write"]], ["read"] if dry_run else ["read", "write"])
        if include_data:
            # Check asset, events, time series, files, and sequences access, relationships, labels, 3D access.
            validator.dataset_data(action, dataset_ids={data_set_id})
        if include_configurations:
            # Check workflow, transformations, extraction pipeline access
            validator.dataset_configurations(action, dataset_ids={data_set_id})

        to_delete: list[ToDelete] = self._create_to_delete_list_purge_dataset(
            client,
            include_data,
            include_configurations,
            selected_data_set_external_id,
            asset_recursive,
        )
        if dry_run:
            results = DeployResults([], "purge", dry_run=True)
            for item in to_delete:
                results[item.display_name] = ResourceDeployResult(item.display_name, deleted=item.total)
        else:
            results = self._delete_resources(to_delete, client, verbose, None, selected_data_set_external_id)
        print(results.counts_table(exclude_columns={"Created", "Changed", "Total"}))
        if archive_dataset and not dry_run:
            self._archive_dataset(client, selected_data_set_external_id)
        return results

    @staticmethod
    def _archive_dataset(client: ToolkitClient, data_set: str) -> None:
        archived = (
            DataSetUpdate(external_id=data_set)
            .external_id.set(str(uuid.uuid4()))
            .metadata.add({"archived": "true"})
            .write_protected.set(True)
        )
        client.data_sets.update(archived)
        print(f"DataSet {data_set} archived")

    @staticmethod
    def _create_to_delete_list_purge_dataset(
        client: ToolkitClient,
        include_data: bool,
        include_configurations: bool,
        data_set_external_id: str,
        asset_recursive: bool,
    ) -> list[ToDelete]:
        config = client.config
        to_delete: list[ToDelete] = []

        if include_data:
            three_d_crud = ThreeDModelCRUD.create_loader(client)
            to_delete.extend(
                [
                    ExternalIdToDelete(
                        RelationshipCRUD.create_loader(client),
                        RelationshipAggregator(client).count(data_set_external_id=data_set_external_id),
                        config.create_api_url("/relationships/delete"),
                    ),
                    IdResourceToDelete(
                        EventCRUD.create_loader(client),
                        EventAggregator(client).count(data_set_external_id=data_set_external_id),
                        config.create_api_url("/events/delete"),
                    ),
                    IdResourceToDelete(
                        FileMetadataCRUD.create_loader(client),
                        FileAggregator(client).count(data_set_external_id=data_set_external_id),
                        config.create_api_url("/files/delete"),
                    ),
                    IdResourceToDelete(
                        TimeSeriesCRUD.create_loader(client),
                        TimeSeriesAggregator(client).count(data_set_external_id=data_set_external_id),
                        config.create_api_url("/timeseries/delete"),
                    ),
                    IdResourceToDelete(
                        SequenceCRUD.create_loader(client),
                        SequenceAggregator(client).count(data_set_external_id=data_set_external_id),
                        config.create_api_url("/sequences/delete"),
                    ),
                    IdResourceToDelete(
                        three_d_crud,
                        sum(1 for _ in three_d_crud.iterate(data_set_external_id=data_set_external_id)),
                        config.create_api_url("/3d/models/delete"),
                    ),
                    AssetToDelete(
                        AssetCRUD.create_loader(client),
                        AssetAggregator(client).count(data_set_external_id=data_set_external_id),
                        config.create_api_url("/assets/delete"),
                        recursive=asset_recursive,
                    ),
                    ExternalIdToDelete(
                        LabelCRUD.create_loader(client),
                        LabelCountAggregator(client).count(data_set_external_id=data_set_external_id),
                        config.create_api_url("/labels/delete"),
                    ),
                ]
            )
        if include_configurations:
            transformation_crud = TransformationCRUD.create_loader(client)
            workflow_crud = WorkflowCRUD.create_loader(client)
            extraction_pipeline_crud = ExtractionPipelineCRUD.create_loader(client)

            to_delete.extend(
                [
                    IdResourceToDelete(
                        transformation_crud,
                        sum(1 for _ in transformation_crud.iterate(data_set_external_id=data_set_external_id)),
                        config.create_api_url("/transformations/delete"),
                    ),
                    ExternalIdToDelete(
                        workflow_crud,
                        sum(1 for _ in workflow_crud.iterate(data_set_external_id=data_set_external_id)),
                        config.create_api_url("/workflows/delete"),
                    ),
                    IdResourceToDelete(
                        extraction_pipeline_crud,
                        sum(1 for _ in extraction_pipeline_crud.iterate(data_set_external_id=data_set_external_id)),
                        config.create_api_url("/extpipes/delete"),
                    ),
                ]
            )
        return to_delete

    @staticmethod
    def _print_panel(resource_type: str, resource: str) -> None:
        print(
            Panel(
                f"[red]WARNING:[/red] This operation [bold]cannot be undone[/bold]! "
                f"Resources in {resource!r} are permanently deleted",
                style="bold",
                title=f"Purge {resource_type}",
                title_align="left",
                border_style="red",
                expand=False,
            )
        )

    def instances(
        self,
        client: ToolkitClient,
        selector: InstanceSelector,
        dry_run: bool = False,
        auto_yes: bool = False,
        unlink: bool = True,
        verbose: bool = False,
    ) -> DeleteResults:
        """Purge instances"""
        io = InstanceIO(client)
        console = Console()
        validator = ValidateAccess(client, default_operation="purge")
        self.validate_instance_access(validator, selector.get_instance_spaces())
        if unlink:
            self.validate_timeseries_access(validator)
            self.validate_file_access(validator)

        total = io.count(selector)
        if total is None or total == 0:
            print("No instances found.")
            return DeleteResults()
        if not dry_run:
            self._print_panel("instances", str(selector))
            if not auto_yes:
                confirm = questionary.confirm(
                    f"Are you really sure you want to purge all {total:,} instances in {selector!s}?",
                    default=False,
                ).ask()
                if not confirm:
                    return DeleteResults()

        process: Callable[[list[InstanceId]], list[dict[str, JsonVal]]] = self._prepare
        if unlink:
            process = partial(self._unlink_prepare, client=client, dry_run=dry_run, console=console, verbose=verbose)

        iteration_count = int(total // io.CHUNK_SIZE + (1 if total % io.CHUNK_SIZE > 0 else 0))
        with HTTPClient(config=client.config) as delete_client:
            process_str = "Would be unlinking" if dry_run else "Unlinking"
            write_str = "Would be deleting" if dry_run else "Deleting"
            results = DeleteResults()
            executor = ProducerWorkerExecutor[list[InstanceId], list[dict[str, JsonVal]]](
                download_iterable=io.download_ids(selector),
                process=process,
                write=partial(self._delete_instance_ids, dry_run=dry_run, delete_client=delete_client, results=results),
                iteration_count=iteration_count,
                max_queue_size=10,
                download_description=f"Retrieving instances from {selector!s}",
                process_description=f"{process_str} instances from files/timeseries" if unlink else "",
                write_description=f"{write_str} instances",
                console=console,
            )

            executor.run()
            executor.raise_on_error()

        prefix = "Would have purged" if dry_run else "Purged"
        if results.failed == 0:
            console.print(f"{prefix} {results.deleted:,} instances in {selector!s}")
        else:
            console.print(
                f"{prefix} {results.deleted:,} instances in {selector!s}, but failed to purge {results.failed:,} instances"
            )
        return results

    def validate_instance_access(self, validator: ValidateAccess, instance_spaces: list[str] | None) -> None:
        space_ids = validator.instances(
            ["read", "write"], spaces=set(instance_spaces) if instance_spaces else None, operation="purge"
        )
        if space_ids is None:
            # Full access
            return
        self.warn(
            LimitedAccessWarning(
                f"You can only purge instances in the following instance spaces: {humanize_collection(space_ids)}."
            )
        )
        return

    def validate_model_access(self, validator: ValidateAccess, view: list[str] | None) -> None:
        space = view[0] if isinstance(view, list) and view and isinstance(view[0], str) else None
        if space_ids := validator.data_model(["read"], spaces={space} if space else None):
            self.warn(
                LimitedAccessWarning(
                    f"You can only select views in the {len(space_ids)} spaces you have access to: {humanize_collection(space_ids)}."
                )
            )

    def validate_timeseries_access(self, validator: ValidateAccess) -> None:
        try:
            ids_by_scope = validator.timeseries(["read", "write"], operation="unlink")
        except AuthorizationError as e:
            self.warn(HighSeverityWarning(f"You cannot unlink time series. You need read and write access: {e!s}"))
            return
        if ids_by_scope is None:
            return
        scope_str = humanize_collection(
            [f"{scope_name} ({humanize_collection(ids)})" for scope_name, ids in ids_by_scope.items()],
            bind_word="and",
        )
        self.warn(LimitedAccessWarning(f"You can only unlink time series in the following scopes: {scope_str}."))

    def validate_file_access(self, validator: ValidateAccess) -> None:
        try:
            ids_by_scope = validator.files(["read", "write"], operation="unlink")
        except AuthorizationError as e:
            self.warn(HighSeverityWarning(f"You cannot unlink files. You need read and write access: {e!s}"))
            return
        if ids_by_scope is None:
            return
        scope_str = humanize_collection(
            [f"{scope_name} ({humanize_collection(ids)})" for scope_name, ids in ids_by_scope.items()],
            bind_word="and",
        )
        self.warn(LimitedAccessWarning(f"You can only unlink files in the following scopes: {scope_str}."))

    @staticmethod
    def _prepare(instance_ids: list[InstanceId]) -> list[dict[str, JsonVal]]:
        output: list[dict[str, JsonVal]] = []
        for instance_id in instance_ids:
            dumped = instance_id.dump(include_instance_type=True)
            # The PySDK uses 'type' instead of 'instanceType' which is required by the delete endpoint
            dumped["instanceType"] = dumped.pop("type")
            # MyPy does not understand that InstanceId.dump() returns dict[str, JsonVal]
            output.append(dumped)  # type: ignore[arg-type]

        return output

    def _unlink_prepare(
        self,
        instance_ids: list[InstanceId],
        client: ToolkitClient,
        dry_run: bool,
        console: Console,
        verbose: bool = False,
    ) -> list[dict[str, JsonVal]]:
        self._unlink_timeseries(instance_ids, client, dry_run, console, verbose)
        self._unlink_files(instance_ids, client, dry_run, console, verbose)
        return self._prepare(instance_ids)

    @staticmethod
    def _delete_instance_ids(
        items: list[dict[str, JsonVal]], dry_run: bool, delete_client: HTTPClient, results: DeleteResults
    ) -> None:
        if dry_run:
            results.deleted += len(items)
            return

        responses = delete_client.request_with_retries(
            ItemsRequest(
                delete_client.config.create_api_url("/models/instances/delete"),
                method="POST",
                # MyPy does not understand that InstanceId.load handles dict[str, JsonVal]
                items=[DeleteItem(item=item, as_id_fun=InstanceId.load) for item in items],  # type: ignore[arg-type]
            )
        )
        for response in responses:
            if isinstance(response, SuccessResponseItems):
                results.deleted += len(response.ids)
            elif isinstance(response, FailedResponseItems | FailedRequestItems):
                results.failed += len(response.ids)
            else:
                results.failed += len(items)

    @staticmethod
    def _unlink_timeseries(
        instances: list[InstanceId], client: ToolkitClient, dry_run: bool, console: Console, verbose: bool
    ) -> list[InstanceId]:
        node_ids = [instance for instance in instances if isinstance(instance, NodeId)]
        if node_ids:
            timeseries = client.time_series.retrieve_multiple(instance_ids=node_ids, ignore_unknown_ids=True)
            if not dry_run and timeseries:
                migrated_timeseries_ids = [ts.id for ts in timeseries if ts.instance_id and ts.pending_instance_id]  # type: ignore[attr-defined]
                client.time_series.unlink_instance_ids(id=migrated_timeseries_ids)
                if verbose:
                    console.print(f"Unlinked {len(migrated_timeseries_ids)} timeseries from datapoints.")
            elif verbose and timeseries:
                console.print(f"Would have unlinked {len(timeseries)} timeseries from datapoints.")
        return instances

    @staticmethod
    def _unlink_files(
        instances: list[InstanceId], client: ToolkitClient, dry_run: bool, console: Console, verbose: bool
    ) -> list[InstanceId]:
        file_ids = [instance for instance in instances if isinstance(instance, NodeId)]
        if file_ids:
            files = client.files.retrieve_multiple(instance_ids=file_ids, ignore_unknown_ids=True)
            if not dry_run and files:
                migrated_file_ids = [
                    file.id
                    for file in files
                    if file.instance_id and file.pending_instance_id and file.id is not None  # type: ignore[attr-defined]
                ]
                client.files.unlink_instance_ids(id=migrated_file_ids)
                if verbose:
                    console.print(f"Unlinked {len(migrated_file_ids)} files from nodes.")
            elif verbose and files:
                console.print(f"Would have unlinked {len(files)} files from their blob content.")
        return instances



import json
from collections import Counter
from pathlib import Path
from unittest.mock import MagicMock

import pytest
from _pytest.monkeypatch import MonkeyPatch
from cognite.client import data_modeling as dm
from cognite.client.data_classes import (
    DataSet,
    DataSetList,
    ExtractionPipeline,
    ExtractionPipelineList,
    FileMetadataList,
    Function,
    FunctionList,
    Group,
    GroupList,
    Transformation,
    TransformationList,
    TransformationScheduleList,
)
from cognite.client.data_classes.agents import Agent, AgentList, AskDocumentAgentTool
from cognite.client.data_classes.aggregations import UniqueResult, UniqueResultList
from cognite.client.data_classes.capabilities import (
    TimeSeriesAcl,
)
from cognite.client.data_classes.data_modeling.statistics import SpaceStatistics
from cognite.client.data_classes.functions import FunctionsStatus
from cognite.client.exceptions import CogniteAPIError
from questionary import Choice
from rich.console import Console

from cognite_toolkit._cdf_tk.client.data_classes.location_filters import LocationFilter, LocationFilterList
from cognite_toolkit._cdf_tk.client.data_classes.search_config import SearchConfig, SearchConfigList, ViewId
from cognite_toolkit._cdf_tk.client.data_classes.streamlit_ import Streamlit, StreamlitList
from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.commands.dump_resource import (
    AgentFinder,
    DataModelFinder,
    DataSetFinder,
    DumpResourceCommand,
    ExtractionPipelineFinder,
    FunctionFinder,
    GroupFinder,
    LocationFilterFinder,
    SearchConfigFinder,
    SpaceFinder,
    StreamlitFinder,
    TransformationFinder,
)
from cognite_toolkit._cdf_tk.cruds import (
    AgentCRUD,
    DataSetsCRUD,
    ExtractionPipelineCRUD,
    FunctionCRUD,
    GroupAllScopedCRUD,
    LocationFilterCRUD,
    SearchConfigCRUD,
    SpaceCRUD,
    StreamlitCRUD,
    TransformationCRUD,
)
from cognite_toolkit._cdf_tk.utils import read_yaml_file
from tests.test_unit.approval_client import ApprovalToolkitClient
from tests.test_unit.utils import MockQuestionary


@pytest.fixture()
def three_transformations() -> TransformationList:
    return TransformationList(
        [
            Transformation(
                1,
                "transformationA",
                name="My First Transformation",
                ignore_null_fields=True,
                created_time=1,
                updated_time=1,
            ),
            Transformation(
                2,
                "transformationB",
                name="My Second Transformation",
                ignore_null_fields=True,
                created_time=1,
                updated_time=1,
            ),
            Transformation(
                3,
                "transformationC",
                name="My Third Transformation",
                ignore_null_fields=True,
                created_time=1,
                updated_time=1,
            ),
        ]
    )


class TestTransformationFinder:
    def test_select_transformations(self, three_transformations: TransformationList, monkeypatch: MonkeyPatch) -> None:
        def select_transformations(choices: list[Choice]) -> list[str]:
            assert len(choices) == len(three_transformations)
            return [choices[1].value, choices[2].value]

        answers = [select_transformations]

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(TransformationFinder.__module__, monkeypatch, answers),
        ):
            client.transformations.list.return_value = three_transformations
            finder = TransformationFinder(client, None)
            selected = finder._interactive_select()

        assert selected == ("transformationB", "transformationC")


class TestDumpTransformations:
    def test_dump_transformations(self, three_transformations: TransformationList, tmp_path: Path) -> None:
        with monkeypatch_toolkit_client() as client:
            client.transformations.retrieve_multiple.return_value = three_transformations[1:]
            client.transformations.schedules.retrieve.return_value = TransformationScheduleList([])

            cmd = DumpResourceCommand(silent=True)
            cmd.dump_to_yamls(
                TransformationFinder(client, ("transformationB", "transformationC")),
                output_dir=tmp_path,
                clean=False,
                verbose=False,
            )
            loader = TransformationCRUD(client, None, None)

        filepaths = list(loader.find_files(tmp_path))
        assert len(filepaths) == 2
        items = [read_yaml_file(filepath) for filepath in filepaths]
        assert items == [loader.dump_resource(t) for t in three_transformations[1:]]


@pytest.fixture()
def three_data_models() -> dm.DataModelList[dm.ViewId]:
    """Three data models in one space. The first two are different versions of the same model."""
    default_args = dict(
        is_global=False,
        last_updated_time=1,
        created_time=1,
        description=None,
        name=None,
    )
    return dm.DataModelList[dm.ViewId](
        [
            dm.DataModel[dm.ViewId](
                "my_space", "my_model", "v1", views=[dm.ViewId("my_space", "firstView", "v1")], **default_args
            ),
            dm.DataModel[dm.ViewId](
                "my_space",
                "my_model",
                "v2",
                views=[
                    dm.ViewId("my_space", "firstView", "v2"),
                    dm.ViewId("my_space2", "secondView", "v2"),
                ],
                **default_args,
            ),
            dm.DataModel[dm.ViewId](
                "my_space", "other_model", "v1", views=[dm.ViewId("my_space", "otherView", "v1")], **default_args
            ),
        ]
    )


class TestDataModelFinder:
    def test_select_data_model(self, toolkit_client_approval: ApprovalToolkitClient, monkeypatch: MonkeyPatch) -> None:
        default_args = dict(
            is_global=False,
            last_updated_time=1,
            created_time=1,
            description=None,
            name=None,
            views=[],
        )
        models = [
            dm.DataModel("my_space", "first_model", "v1", **default_args),
            dm.DataModel("my_space2", "second_model", "v1", **default_args),
        ]
        toolkit_client_approval.append(dm.DataModel, models)
        selected = models[1].as_id()
        finder = DataModelFinder(toolkit_client_approval.mock_client, None)
        answers = ["my_space2", selected, False]
        with MockQuestionary(DataModelFinder.__module__, monkeypatch, answers):
            result = finder._interactive_select()

        assert result == selected
        assert finder.data_model.as_id() == selected

    def test_select_data_model_multiple_versions(
        self, three_data_models: dm.DataModelList[dm.ViewId], monkeypatch: MonkeyPatch
    ) -> None:
        def select_data_model(choices: list[Choice]) -> dm.DataModelId:
            assert len(choices) == 2
            return choices[0].value

        def select_version_of_model(choices: list[Choice]) -> str:
            assert len(choices) == 2

            return choices[0].value

        answers = [select_data_model, True, select_version_of_model]

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(DataModelFinder.__module__, monkeypatch, answers),
        ):
            # Last two are different models
            client.data_modeling.data_models.list.return_value = three_data_models[1:]
            # First two are different versions of the same model
            client.data_modeling.data_models.retrieve.return_value = three_data_models[:2]

            finder = DataModelFinder(client, None)
            selected = finder._interactive_select()

        assert selected == three_data_models[0].as_id()


@pytest.fixture()
def three_location_filters() -> LocationFilterList:
    return LocationFilterList(
        [
            LocationFilter(1, external_id="filterA", name="Filter A", created_time=1, updated_time=1),
            LocationFilter(2, external_id="filterB", name="Filter B", created_time=1, updated_time=1),
            LocationFilter(3, external_id="filterC", name="Filter C", created_time=1, updated_time=1),
        ]
    )


class TestLocationFilterFinder:
    def test_select_location_filter(self, three_location_filters: LocationFilterList, monkeypatch: MonkeyPatch) -> None:
        def select_filters(choices: list[Choice]) -> list[str]:
            assert len(choices) == len(three_location_filters)
            return [choices[1].value, choices[2].value]

        answers = [select_filters]

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(LocationFilterFinder.__module__, monkeypatch, answers),
        ):
            client.search.locations.list.return_value = three_location_filters
            finder = LocationFilterFinder(client, None)
            selected = finder._interactive_select()

        assert selected == ("filterB", "filterC")


class TestDumpLocationFilter:
    def test_dump_location_filter(self, three_location_filters: LocationFilterList, tmp_path: Path) -> None:
        with monkeypatch_toolkit_client() as client:
            client.search.locations.list.return_value = three_location_filters
            cmd = DumpResourceCommand(silent=True)
            cmd.dump_to_yamls(
                LocationFilterFinder(client, ("filterB", "filterC")),
                output_dir=tmp_path,
                clean=False,
                verbose=False,
            )
            loader = LocationFilterCRUD(client, None, None)

        filepaths = list(loader.find_files(tmp_path))
        assert len(filepaths) == 2
        items = [item for filepath in filepaths for item in loader.load_resource_file(filepath)]
        assert items == three_location_filters[1:].as_write().dump()


@pytest.fixture()
def three_agents() -> AgentList:
    return AgentList(
        [
            Agent(
                external_id=f"agent{character}",
                name=f"Agent {character}",
                description=f"This is Agent {character}",
                model="gpt-3.5-turbo",
                tools=[
                    AskDocumentAgentTool(
                        name=f"tool{character}",
                        description=f"This is tool {character}",
                    )
                ],
            )
            for character in ["A", "B", "C"]
        ]
    )


class TestDumpAgents:
    def test_dump_agents(self, three_agents: AgentList, tmp_path: Path) -> None:
        with monkeypatch_toolkit_client() as client:
            client.agents.retrieve.return_value = three_agents[1:]
            cmd = DumpResourceCommand(silent=True)
            cmd.dump_to_yamls(
                AgentFinder(client, tuple([agent.external_id for agent in three_agents[1:]])),
                output_dir=tmp_path,
                clean=False,
                verbose=False,
            )
            loader = AgentCRUD(client, None, None)

        filepaths = list(loader.find_files(tmp_path))
        assert len(filepaths) == 2
        items = [read_yaml_file(filepath) for filepath in filepaths]
        assert items == [loader.dump_resource(agent) for agent in three_agents[1:]]

    def test_interactive_select_agents(self, three_agents: AgentList, monkeypatch: MonkeyPatch) -> None:
        def select_agents(choices: list[Choice]) -> list[str]:
            assert len(choices) == len(three_agents)
            return [choices[1].value, choices[2].value]

        answers = [select_agents]

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(AgentFinder.__module__, monkeypatch, answers),
        ):
            client.agents.list.return_value = three_agents
            finder = AgentFinder(client, None)
            selected = finder._interactive_select()

        assert selected == ("agentB", "agentC")


@pytest.fixture()
def three_extraction_pipelines() -> ExtractionPipelineList:
    return ExtractionPipelineList(
        [
            ExtractionPipeline(
                1, external_id="pipelineA", name="Pipeline A", data_set_id=123, created_time=1, last_updated_time=1
            ),
            ExtractionPipeline(
                2, external_id="pipelineB", name="Pipeline B", data_set_id=123, created_time=1, last_updated_time=1
            ),
            ExtractionPipeline(
                3, external_id="pipelineC", name="Pipeline C", data_set_id=123, created_time=1, last_updated_time=1
            ),
        ]
    )


class TestExtractionPipelineFinder:
    def test_select_extraction_pipelines(
        self, three_extraction_pipelines: ExtractionPipelineList, monkeypatch: MonkeyPatch
    ) -> None:
        def select_pipelines(choices: list[Choice]) -> list[str]:
            assert len(choices) == len(three_extraction_pipelines)
            return [choices[1].value, choices[2].value]

        answers = [select_pipelines]

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(ExtractionPipelineFinder.__module__, monkeypatch, answers),
        ):
            client.extraction_pipelines.list.return_value = three_extraction_pipelines
            finder = ExtractionPipelineFinder(client, None)
            selected = finder._interactive_select()

        assert selected == ("pipelineB", "pipelineC")


class TestDumpExtractionPipeline:
    def test_dump_extraction_pipelines(
        self, three_extraction_pipelines: ExtractionPipelineList, tmp_path: Path
    ) -> None:
        with monkeypatch_toolkit_client() as toolkit_client:
            approval_client = ApprovalToolkitClient(toolkit_client, allow_reverse_lookup=True)
            approval_client.append(ExtractionPipeline, three_extraction_pipelines[1:])
            toolkit_client.extraction_pipelines.config.retrieve.side_effect = CogniteAPIError(
                "There is no config stored for pipeline", code=404
            )

            client = approval_client.mock_client
            cmd = DumpResourceCommand(silent=True)
            cmd.dump_to_yamls(
                ExtractionPipelineFinder(client, ("pipelineB", "pipelineC")),
                output_dir=tmp_path,
                clean=False,
                verbose=False,
            )
            loader = ExtractionPipelineCRUD(client, None, None)

            filepaths = list(loader.find_files(tmp_path))
            items = sorted(
                [read_yaml_file(filepath) for filepath in filepaths],
                key=lambda d: d.get("external_id", d.get("externalId")),
            )
            expected = sorted(
                [loader.dump_resource(ep) for ep in three_extraction_pipelines[1:]],
                key=lambda d: d.get("external_id", d.get("externalId")),
            )
            assert items == expected


@pytest.fixture()
def three_groups() -> GroupList:
    return GroupList(
        [
            Group(
                "Group A",
                source_id="123",
                capabilities=[TimeSeriesAcl([TimeSeriesAcl.Action.Read], TimeSeriesAcl.Scope.All())],
            ),
            Group(
                "Group B",
                source_id="456",
                capabilities=[TimeSeriesAcl([TimeSeriesAcl.Action.Write], TimeSeriesAcl.Scope.All())],
            ),
            Group(
                "Group C",
                source_id="789",
                capabilities=[
                    TimeSeriesAcl([TimeSeriesAcl.Action.Read, TimeSeriesAcl.Action.Write], TimeSeriesAcl.Scope.All())
                ],
            ),
        ]
    )


class TestGroupFinder:
    def test_select_groups(self, three_groups: GroupList, monkeypatch: MonkeyPatch) -> None:
        def select_groups(choices: list[Choice]) -> list[list[Group]]:
            assert len(choices) == len(three_groups)
            return [choices[1].value, choices[2].value]

        answers = [select_groups]

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(GroupFinder.__module__, monkeypatch, answers),
        ):
            client.iam.groups.list.return_value = three_groups
            finder = GroupFinder(client, None)
            selected = finder._interactive_select()

        assert selected == ("Group B", "Group C")


class TestDumpGroups:
    def test_dump_groups(self, three_groups: GroupList, tmp_path: Path) -> None:
        with monkeypatch_toolkit_client() as client:
            client.iam.groups.list.return_value = three_groups

            cmd = DumpResourceCommand(silent=True)
            cmd.dump_to_yamls(
                GroupFinder(client, ("Group B", "Group C")),
                output_dir=tmp_path,
                clean=False,
                verbose=False,
            )
            loader = GroupAllScopedCRUD(client, None, None)

        filepaths = list(loader.find_files(tmp_path))
        assert len(filepaths) == 2
        items = sorted([read_yaml_file(filepath) for filepath in filepaths], key=lambda d: d.get("name"))
        expected = sorted([loader.dump_resource(group) for group in three_groups[1:]], key=lambda d: d.get("name"))
        assert items == expected


@pytest.fixture()
def three_functions() -> FunctionList:
    return FunctionList(
        [
            Function(
                external_id="functionA",
                name="Function A",
                description="This is Function A",
                created_time=1,
                file_id=123,
            ),
            Function(
                external_id="functionB",
                name="Function B",
                description="This is Function B",
                created_time=1,
                file_id=456,
            ),
            Function(
                external_id="functionC",
                name="Function C",
                description="This is Function C",
                created_time=1,
                file_id=789,
            ),
        ]
    )


class TestDumpFunctions:
    def test_dump_functions(self, three_functions: FunctionList, tmp_path: Path) -> None:
        with monkeypatch_toolkit_client() as client:
            client.functions.retrieve_multiple.return_value = three_functions[1:]
            client.functions.status.return_value = FunctionsStatus("activated")
            client.files.retrieve.return_value = None
            client.files.download_bytes.side_effect = CogniteAPIError(
                "File ids not found",
                code=400,
            )

            cmd = DumpResourceCommand(silent=True)
            cmd.dump_to_yamls(
                FunctionFinder(client, ("functionB", "functionC")),
                output_dir=tmp_path,
                clean=False,
                verbose=False,
            )
            loader = FunctionCRUD(client, None, None)

        filepaths = list(loader.find_files(tmp_path))
        assert len(filepaths) == 2
        items = sorted([read_yaml_file(filepath) for filepath in filepaths], key=lambda d: d["externalId"])
        expected = sorted([loader.dump_resource(func) for func in three_functions[1:]], key=lambda d: d["externalId"])
        assert items == expected

    def test_interactive_select_functions(self, three_functions: FunctionList, monkeypatch: MonkeyPatch) -> None:
        def select_functions(choices: list[Choice]) -> list[str]:
            assert len(choices) == len(three_functions)
            return [choices[1].value, choices[2].value]

        answers = [select_functions]

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(FunctionFinder.__module__, monkeypatch, answers),
        ):
            client.functions.list.return_value = three_functions
            finder = FunctionFinder(client, None)
            selected = finder._interactive_select()

        assert selected == ("functionB", "functionC")


@pytest.fixture()
def three_datasets() -> DataSetList:
    return DataSetList(
        [
            DataSet(external_id="datasetA", name="Dataset A", created_time=1, last_updated_time=2),
            DataSet(external_id="datasetB", name="Dataset B", created_time=1, last_updated_time=2),
            DataSet(external_id="datasetC", name="Dataset C", created_time=1, last_updated_time=2),
        ]
    )


class TestDataSetFinder:
    def test_select_datasets(self, three_datasets: DataSetList, monkeypatch: MonkeyPatch) -> None:
        def select_datasets(choices: list[Choice]) -> list[str]:
            assert len(choices) == len(three_datasets)
            return [choices[1].value, choices[2].value]

        answers = [select_datasets]

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(DataSetFinder.__module__, monkeypatch, answers),
        ):
            client.data_sets.list.return_value = three_datasets
            finder = DataSetFinder(client, None)
            selected = finder._interactive_select()

        assert selected == ("datasetB", "datasetC")


class TestDumpDataSets:
    def test_dump_datasets(self, three_datasets: DataSetList, tmp_path: Path) -> None:
        with monkeypatch_toolkit_client() as client:
            client.data_sets.retrieve_multiple.return_value = three_datasets[1:]

            cmd = DumpResourceCommand(silent=True)
            cmd.dump_to_yamls(
                DataSetFinder(client, ("datasetB", "datasetC")),
                output_dir=tmp_path,
                clean=False,
                verbose=False,
            )
            loader = DataSetsCRUD(client, None, None)

        filepaths = list(loader.find_files(tmp_path))
        assert len(filepaths) == 2
        items = sorted([read_yaml_file(filepath) for filepath in filepaths], key=lambda d: d["externalId"])
        expected = sorted([loader.dump_resource(ds) for ds in three_datasets[1:]], key=lambda d: d["externalId"])
        assert items == expected


@pytest.fixture()
def three_streamlit_apps() -> StreamlitList:
    return StreamlitList(
        [
            Streamlit(
                external_id="appA",
                name="App A",
                description="This is App A",
                created_time=1,
                last_updated_time=1,
                entrypoint="main.py",
                creator="me",
            ),
            Streamlit(
                external_id="appB",
                name="App B",
                description="This is App B",
                created_time=1,
                last_updated_time=1,
                entrypoint="main.py",
                creator="me",
            ),
            Streamlit(
                external_id="appC",
                name="App C",
                description="This is App C",
                created_time=1,
                last_updated_time=1,
                entrypoint="main.py",
                creator="me",
            ),
        ]
    )


class TestDumpStreamlitApps:
    def test_interactive_select_streamlit_apps(
        self, three_streamlit_apps: StreamlitList, monkeypatch: MonkeyPatch
    ) -> None:
        def select_streamlit_apps(choices: list[Choice]) -> list[str]:
            assert len(choices) == len(three_streamlit_apps)
            return [choices[1].value, choices[2].value]

        answers = ["me", select_streamlit_apps]

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(StreamlitFinder.__module__, monkeypatch, answers),
        ):
            client.files.list.return_value = FileMetadataList([app.as_file() for app in three_streamlit_apps])
            counts_by_creator = Counter([app.creator for app in three_streamlit_apps])
            client.documents.aggregate_unique_values.return_value = UniqueResultList(
                [UniqueResult(count=count, values=[creator]) for creator, count in counts_by_creator.items()]
            )
            finder = StreamlitFinder(client, None)
            selected = finder._interactive_select()

        assert selected == ("appB", "appC")

    def test_dump_streamlit_app(self, three_streamlit_apps: StreamlitList, tmp_path: Path) -> None:
        with monkeypatch_toolkit_client() as client:
            client.files.retrieve_multiple.return_value = FileMetadataList([three_streamlit_apps[1].as_file()])
            client.files.download_bytes.return_value = json.dumps(
                {
                    "entrypoint": "main.py",
                    "files": {
                        "main.py": {
                            "content": {
                                "text": 'import streamlit as st\nfrom cognite.client import CogniteClient\n\nst.title("An example app in CDF")\nclient = CogniteClient()\n\n\n@st.cache_data\ndef get_assets():\n    assets = client.assets.list(limit=1000).to_pandas()\n    assets = assets.fillna(0)\n    return assets\n\n\nst.write(get_assets())\n',
                                "$case": "text",
                            }
                        }
                    },
                    "requirements": ["pyodide-http==0.2.1", "cognite-sdk==7.51.1"],
                }
            ).encode("utf-8")

            cmd = DumpResourceCommand(silent=True)
            cmd.dump_to_yamls(
                StreamlitFinder(client, ("appB",)),
                output_dir=tmp_path,
                clean=False,
                verbose=False,
            )
            loader = StreamlitCRUD(client, None, None)

        filepaths = list(loader.find_files(tmp_path))
        assert len(filepaths) == 1
        config_file = filepaths[0]
        loaded = read_yaml_file(config_file)
        expected = loader.dump_resource(three_streamlit_apps[1])
        assert loaded == expected

        app_dir = config_file.parent / "appB"
        assert app_dir.is_dir()
        app_files = list(app_dir.iterdir())
        assert set(app_files) == {
            app_dir / "main.py",
            app_dir / "requirements.txt",
        }

    @pytest.mark.parametrize(
        "content,expected_warning,expected_severity",
        [
            pytest.param(
                None,
                "The source code for 'appB' could not be retrieved from CDF.",
                "high",
                id="File not found",
            ),
            pytest.param(
                '{"not": "json"',
                "The JSON content for the Streamlit app 'appB' is corrupt and could not be "
                "extracted. Download file with the same external id manually to remediate.",
                "high",
                id="Corrupted JSON",
            ),
            pytest.param(
                json.dumps({"requirements": [], "files": {}, "entrypoint": "main.py"}),
                "The Streamlit app 'appB' does not have any files to dump. It is likely corrupted.",
                "high",
                id="No files in JSON",
            ),
            pytest.param(
                json.dumps(
                    {
                        "requirements": ["foo==1.0"],
                        "files": {
                            "page/CDF_demo.py": {"content": {"text": True}},
                            "main.py": {"content": {"text": "print('hi')"}},
                        },
                        "entrypoint": "main.py",
                    }
                ),
                "The Streamlit app 'appB' has a file page/CDF_demo.py with invalid content. Skipping...",
                "high",
                id="File with no content",
            ),
            pytest.param(
                json.dumps(
                    {
                        "requirements": ["foo==1.0"],
                        "files": {"main.py": {"content": {"text": "print('hi')"}}},
                        "entrypoint": "notfound.py",
                    }
                ),
                "The Streamlit app 'appB' has an entry point notfound.py that was not found in the files. The app may be corrupted.",
                "high",
                id="Entrypoint not found",
            ),
        ],
    )
    def test_dump_code(
        self,
        content: str | None,
        expected_warning: str,
        expected_severity: str,
        three_streamlit_apps: StreamlitList,
        tmp_path: Path,
    ) -> None:
        console = MagicMock(spec=Console)
        app = three_streamlit_apps[1]
        with monkeypatch_toolkit_client() as client:
            if content is None:
                client.files.download_bytes.side_effect = CogniteAPIError(
                    message=f"Files ids not found: {app.external_id}", code=400, missing=[app.external_id]
                )
            else:
                client.files.download_bytes.return_value = content.encode("utf-8")

            finder = StreamlitFinder(client, (app.external_id,))
            finder.dump_code(app, tmp_path, console)

            assert console.print.call_count == 1
            severity, actual_message = console.print.call_args.args
            assert expected_warning in actual_message
            assert expected_severity in str(severity).casefold()


@pytest.fixture()
def three_spaces() -> dm.SpaceList:
    return dm.SpaceList(
        [
            dm.Space(
                space="spaceA",
                name="Space A",
                description="This is Space A",
                is_global=False,
                last_updated_time=1,
                created_time=1,
            ),
            dm.Space(
                space="spaceB",
                name="Space B",
                description="This is Space B",
                is_global=False,
                last_updated_time=1,
                created_time=1,
            ),
            dm.Space(
                space="spaceC",
                name="Space C",
                description="This is Space C",
                is_global=False,
                last_updated_time=1,
                created_time=1,
            ),
        ]
    )


class TestSpaceFinder:
    def test_select_spaces(self, three_spaces: dm.SpaceList, monkeypatch: MonkeyPatch) -> None:
        mock_stats_by_space = {
            "spaceA": SpaceStatistics(
                space="spaceA",
                nodes=5,
                edges=3,
                containers=2,
                views=1,
                data_models=1,
                soft_deleted_edges=0,
                soft_deleted_nodes=0,
            ),
            "spaceB": SpaceStatistics(
                space="spaceB",
                nodes=10,
                edges=7,
                containers=3,
                views=2,
                data_models=1,
                soft_deleted_edges=0,
                soft_deleted_nodes=0,
            ),
            "spaceC": SpaceStatistics(
                space="spaceC",
                nodes=15,
                edges=12,
                containers=4,
                views=3,
                data_models=2,
                soft_deleted_edges=0,
                soft_deleted_nodes=0,
            ),
        }

        def select_spaces(choices: list[Choice]) -> list[str]:
            assert len(choices) == len(three_spaces)
            return [choices[1].value, choices[2].value]

        answers = [select_spaces]
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary("cognite_toolkit._cdf_tk.utils.interactive_select", monkeypatch, answers),
        ):
            client.data_modeling.spaces.list.return_value = three_spaces
            monkeypatch.setattr(
                "cognite_toolkit._cdf_tk.utils.interactive_select.DataModelingSelect.stats_by_space",
                mock_stats_by_space,
            )
            finder = SpaceFinder(client, None)
            selected = finder._interactive_select()

        assert selected == ("spaceB", "spaceA")


class TestDumpSpaces:
    def test_dump_spaces(self, three_spaces: dm.SpaceList, tmp_path: Path) -> None:
        with monkeypatch_toolkit_client() as client:
            client.data_modeling.spaces.retrieve.return_value = three_spaces[1:]

            cmd = DumpResourceCommand(silent=True)
            cmd.dump_to_yamls(
                SpaceFinder(client, ("spaceB", "spaceC")),
                output_dir=tmp_path,
                clean=False,
                verbose=False,
            )
            loader = SpaceCRUD(client, None, None)

        filepaths = list(loader.find_files(tmp_path))
        assert len(filepaths) == 2
        items = sorted([read_yaml_file(filepath) for filepath in filepaths], key=lambda d: d["space"])
        expected = sorted([loader.dump_resource(s) for s in three_spaces[1:]], key=lambda d: d["space"])
        assert items == expected


@pytest.fixture()
def three_search_configs() -> SearchConfigList:
    return SearchConfigList(
        [
            SearchConfig(
                view=ViewId(external_id="searchConfigA", space="spaceA"), id=1, created_time=1, updated_time=2
            ),
            SearchConfig(
                view=ViewId(external_id="searchConfigB", space="spaceB"), id=2, created_time=1, updated_time=2
            ),
            SearchConfig(
                view=ViewId(external_id="searchConfigC", space="spaceC"), id=3, created_time=1, updated_time=2
            ),
        ]
    )


class TestSearchConfigFinder:
    def test_select_search_configs(self, three_search_configs: SearchConfigList, monkeypatch: MonkeyPatch) -> None:
        def select_search_configs(choices: list[Choice]) -> list[str]:
            assert len(choices) == len(three_search_configs)
            return [choices[1].value, choices[2].value]

        answers = [select_search_configs]

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(SearchConfigFinder.__module__, monkeypatch, answers),
        ):
            client.search.configurations.list.return_value = three_search_configs
            finder = SearchConfigFinder(client, None)
            selected = finder._interactive_select()

        assert selected == (
            ViewId(external_id="searchConfigB", space="spaceB"),
            ViewId(external_id="searchConfigC", space="spaceC"),
        )


class TestDumpSearchConfigs:
    def test_dump_search_configs(self, three_search_configs: SearchConfigList, tmp_path: Path) -> None:
        with monkeypatch_toolkit_client() as client:
            client.search.configurations.list.return_value = three_search_configs[1:]

            cmd = DumpResourceCommand(silent=True)
            cmd.dump_to_yamls(
                SearchConfigFinder(
                    client,
                    (
                        ViewId(external_id="searchConfigB", space="spaceB"),
                        ViewId(external_id="searchConfigC", space="spaceC"),
                    ),
                ),
                output_dir=tmp_path,
                clean=False,
                verbose=False,
            )
            loader = SearchConfigCRUD(client, None, None)

        filepaths = list(loader.find_files(tmp_path))
        assert len(filepaths) == 2
        items = sorted([read_yaml_file(filepath) for filepath in filepaths], key=lambda d: d["view"]["externalId"])
        expected = sorted(
            [loader.dump_resource(sc) for sc in three_search_configs[1:]], key=lambda d: d["view"]["externalId"]
        )
        assert items == expected



from pydantic import Field

from cognite_toolkit._cdf_tk.constants import (
    DM_EXTERNAL_ID_PATTERN,
    DM_VERSION_PATTERN,
    SPACE_FORMAT_PATTERN,
)

from .base import ToolkitResource


class GraphQLDataModelYAML(ToolkitResource):
    space: str = Field(
        description="The workspace for the data model, a unique identifier for the space.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    external_id: str = Field(
        description="External-id of the data model.",
        min_length=1,
        max_length=255,
        pattern=DM_EXTERNAL_ID_PATTERN,
    )
    version: str = Field(
        description="Version of the data model.",
        max_length=43,
        pattern=DM_VERSION_PATTERN,
    )
    name: str | None = Field(
        default=None,
        description="Human readable name for the data model.",
        max_length=255,
    )
    description: str | None = Field(
        default=None,
        description="Description of the data model.",
        max_length=1024,
    )
    dml: str | None = Field(
        default=None,
        description="The path to the graphql file containing the DML schema. If not provided, the file"
        "is assumed to be named <config_file>.graphql in the same directory as the config file.",
    )
    previous_version: str | None = Field(
        default=None,
        max_length=43,
        pattern=DM_VERSION_PATTERN,
    )
    preserve_dml: bool | None = None



import json

from cognite_toolkit._cdf_tk.client import ToolkitClientConfig
from cognite_toolkit._cdf_tk.utils.http_client import HTTPClient, ParamRequest, SimpleBodyRequest, SuccessResponse


class TestHttpClient:
    def test_get_request(self, toolkit_client_config: ToolkitClientConfig) -> None:
        config = toolkit_client_config
        with HTTPClient(config) as client:
            result = client.request(
                ParamRequest(
                    endpoint_url=config.create_api_url(""),
                    method="GET",
                )
            )

        assert len(result) == 1
        response = result[0]
        assert isinstance(response, SuccessResponse)
        assert response.status_code == 200

    def test_post_request(self, toolkit_client_config: ToolkitClientConfig) -> None:
        config = toolkit_client_config
        with HTTPClient(config) as client:
            result = client.request(
                SimpleBodyRequest(
                    endpoint_url=config.create_api_url("timeseries/list"), method="POST", body_content={"limit": 1}
                )
            )
        assert len(result) == 1
        response = result[0]
        assert isinstance(response, SuccessResponse)
        assert response.status_code == 200
        response_body = json.loads(response.body)
        assert "items" in response_body
        assert isinstance(response_body["items"], list)
        assert len(response_body["items"]) <= 1



from __future__ import annotations

from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from cognite_toolkit._cdf_tk.commands import BuildCommand
from cognite_toolkit._cdf_tk.cruds import CRUDS_BY_FOLDER_NAME
from cognite_toolkit._cdf_tk.data_classes import BuildConfigYAML
from cognite_toolkit._cdf_tk.utils import iterate_modules
from tests.data import PROJECT_FOR_TEST
from tests.test_unit.test_cdf_tk.constants import BUILD_DIR


class TestBuildConfigYAML:
    def test_build_config_create_valid_build_folder(self, project_for_test_config_dev_yaml: str) -> None:
        build_env_name = "dev"
        cdf_toml = CDFToml.load(PROJECT_FOR_TEST)
        config = BuildConfigYAML.load_from_directory(PROJECT_FOR_TEST, build_env_name)
        available_modules = {module.name for module, _ in iterate_modules(PROJECT_FOR_TEST)}
        config.environment.selected = list(available_modules)

        BuildCommand(silent=True).build_config(
            BUILD_DIR, PROJECT_FOR_TEST, config=config, packages=cdf_toml.modules.packages, clean=True, verbose=False
        )

        # The resulting build folder should only have subfolders that are matching the folder name
        # used by the loaders.
        invalid_resource_folders = [
            dir_.name for dir_ in BUILD_DIR.iterdir() if dir_.is_dir() and dir_.name not in CRUDS_BY_FOLDER_NAME
        ]
        assert not invalid_resource_folders, f"Invalid resource folders after build: {invalid_resource_folders}"



import sys
import tempfile
import time
from collections.abc import Hashable, Iterator
from dataclasses import dataclass
from pathlib import Path
from threading import RLock
from typing import Any, Literal, overload
from urllib.parse import urlparse

from cognite.client.credentials import OAuthClientCredentials
from cognite.client.data_classes import (
    ClientCredentials,
    OidcCredentials,
)
from cognite.client.data_classes.data_modeling import Edge, Node, ViewId
from cognite.client.data_classes.filters import SpaceFilter
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils.useful_types import SequenceNotStr
from filelock import BaseFileLock, FileLock, Timeout
from rich.console import Console

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.data_classes.raw import RawTable
from cognite_toolkit._cdf_tk.constants import ENV_VAR_PATTERN, MAX_ROW_ITERATION_RUN_QUERY, MAX_RUN_QUERY_FREQUENCY_MIN
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitError,
    ToolkitRequiredValueError,
    ToolkitThrottledError,
    ToolkitTypeError,
    ToolkitValueError,
)
from cognite_toolkit._cdf_tk.tk_warnings import (
    HighSeverityWarning,
    MediumSeverityWarning,
)
from cognite_toolkit._cdf_tk.utils import humanize_collection

from .file import sanitize_filename
from .sql_parser import SQLParser

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self


def try_find_error(credentials: OidcCredentials | ClientCredentials | None) -> str | None:
    if credentials is None:
        return None
    missing: list[str] = []
    if client_id := ENV_VAR_PATTERN.match(credentials.client_id):
        missing.append(client_id.group(1))
    if client_secret := ENV_VAR_PATTERN.match(credentials.client_secret):
        missing.append(client_secret.group(1))
    if missing:
        plural = "s are" if len(missing) > 1 else " is"
        return f"The environment variable{plural} not set: {humanize_collection(missing)}."
    if isinstance(credentials, ClientCredentials):
        return None
    try:
        result = urlparse(credentials.token_uri)
        if not all([result.scheme, result.netloc]):
            raise ValueError
    except ValueError:
        return f"The tokenUri {credentials.token_uri!r} is not a valid URI."
    return None


@overload
def iterate_instances(
    client: ToolkitClient,
    instance_type: Literal["node"] = "node",
    space: str | SequenceNotStr[str] | None = None,
    source: ViewId | None = None,
    console: Console | None = None,
) -> Iterator[Node]: ...


@overload
def iterate_instances(
    client: ToolkitClient,
    instance_type: Literal["edge"],
    space: str | SequenceNotStr[str] | None = None,
    source: ViewId | None = None,
    console: Console | None = None,
) -> Iterator[Edge]: ...


def iterate_instances(
    client: ToolkitClient,
    instance_type: Literal["node", "edge"] = "node",
    space: str | SequenceNotStr[str] | None = None,
    source: ViewId | None = None,
    console: Console | None = None,
) -> Iterator[Node] | Iterator[Edge]:
    """Toolkit specific implementation of the client.data_modeling.instances(...) method to account for 408.

    In addition, we enforce sort based on the argument below (provided by Alex B.).
    """
    body: dict[str, Any] = {"limit": 1_000, "cursor": None, "instanceType": instance_type}
    # Without a sort, the sort is implicitly by the internal id, as cursoring needs a stable sort.
    # By making the sort be on external_id, Postgres should pick the index that's on (project_id, space, external_id)
    # WHERE deleted_at IS NULL. In other words, avoiding soft deleted instances.
    body["sort"] = [
        {
            "property": [instance_type, "space"],
            "direction": "ascending",
        },
        {
            "property": [instance_type, "externalId"],
            "direction": "ascending",
        },
    ]
    url = f"/api/{client._API_VERSION}/projects/{client.config.project}/models/instances/list"
    if space:
        body["filter"] = SpaceFilter(space=space, instance_type=instance_type).dump()
    if source:
        body["sources"] = [{"source": source.dump(include_type=True, camel_case=True)}]
    while True:
        try:
            response = client.post(url=url, json=body)
        except CogniteAPIError as e:
            if e.code == 408 and body["limit"] > 1:
                MediumSeverityWarning(
                    f"Timeout with limit {body['limit']}, retrying with {body['limit'] // 2}."
                ).print_warning(include_timestamp=True, console=console)
                body["limit"] = body["limit"] // 2
                continue
            raise e
        response_body = response.json()
        if instance_type == "node":
            yield from (Node.load(node) for node in response_body["items"])
        else:
            yield from (Edge.load(edge) for edge in response_body["items"])
        next_cursor = response_body.get("nextCursor")
        if next_cursor is None:
            break
        body["cursor"] = next_cursor


@overload
def read_auth(
    authentication: Any,
    client_config: ToolkitClientConfig,
    identifier: Hashable,
    resource_name: str,
    allow_oidc: Literal[False] = False,
    console: Console | None = None,
) -> ClientCredentials: ...


@overload
def read_auth(
    authentication: Any,
    client_config: ToolkitClientConfig,
    identifier: Hashable,
    resource_name: str,
    allow_oidc: Literal[True],
    console: Console | None = None,
) -> ClientCredentials | OidcCredentials: ...


def read_auth(
    authentication: Any,
    client_config: ToolkitClientConfig,
    identifier: Hashable,
    resource_name: str,
    allow_oidc: bool = False,
    console: Console | None = None,
) -> ClientCredentials | OidcCredentials:
    if authentication is None:
        if client_config.is_strict_validation or not isinstance(client_config.credentials, OAuthClientCredentials):
            raise ToolkitRequiredValueError(f"Authentication is missing for {resource_name} {identifier!r}.")
        else:
            HighSeverityWarning(
                f"Authentication is missing for {resource_name} {identifier!r}. Falling back to the Toolkit credentials"
            ).print_warning(console=console)
        return ClientCredentials(client_config.credentials.client_id, client_config.credentials.client_secret)
    elif not isinstance(authentication, dict):
        raise ToolkitTypeError(f"Authentication must be a dictionary for {resource_name} {identifier!r}")
    elif "clientId" not in authentication or "clientSecret" not in authentication:
        raise ToolkitRequiredValueError(
            f"Authentication must contain clientId and clientSecret for {resource_name} {identifier!r}"
        )
    elif allow_oidc and "tokenUri" in authentication and "cdfProjectName" in authentication:
        return OidcCredentials.load(authentication)
    else:
        return ClientCredentials(authentication["clientId"], authentication["clientSecret"])


def get_transformation_sources(query: str) -> list[RawTable | str]:
    """Search the SQL query for source tables."""
    parser = SQLParser(query, operation="Lookup transformation source")
    parser.parse()

    tables: list[RawTable | str] = []
    for table in parser.sources:
        if table.schema == "_cdf":
            tables.append(table.name)
        else:
            tables.append(RawTable(db_name=table.schema, table_name=table.name))
    return tables


def get_transformation_destination_columns(query: str) -> list[str]:
    """Search the SQL query for destination columns."""
    parser = SQLParser(query, operation="Lookup transformation destination columns")
    parser.parse()
    return parser.destination_columns


def metadata_key_counts(
    client: ToolkitClient,
    resource: Literal["assets", "events", "files", "timeseries", "sequences"],
    data_sets: list[int] | None = None,
    hierarchies: list[int] | None = None,
) -> list[tuple[str, int]]:
    """Get the metadata key counts for a given resource.

    Args:
        client: ToolkitClient instance
        resource: The resource to get the metadata key counts for. Can be one of "assets", "events", "files", "timeseries", or "sequences".
        data_sets: A list of data set IDs to filter by. If None, no filtering is applied.
        hierarchies: A list of hierarchy IDs to filter by. If None, no filtering is applied.

    Returns:
        A dictionary with the metadata keys as keys and the counts as values.
    """
    if hierarchies and resource != "assets":
        raise ToolkitValueError(
            f"Hierarchies filtering for metadata keys are only supported for assets, but {resource} was provided."
        )
    where_clause = _create_where_clause(data_sets, hierarchies)

    query = f"""WITH meta AS (
         SELECT cast_to_strings(metadata) AS metadata_array
         FROM _cdf.{resource}{where_clause}
       ),
       exploded AS (
         SELECT explode(metadata_array) AS json_str
         FROM meta
       ),
       parsed AS (
         SELECT from_json(json_str, 'map<string,string>') AS json_map
         FROM exploded
       ),
       keys_extracted AS (
         SELECT map_keys(json_map) AS keys_array
         FROM parsed
       ),
       all_keys AS (
         SELECT explode(keys_array) AS key
         FROM keys_extracted
       )
       SELECT key, COUNT(key) AS key_count
       FROM all_keys
       GROUP BY key
       ORDER BY key_count DESC;
"""
    results = client.transformations.preview(query, convert_to_string=False, limit=None, source_limit=None)
    return [(item["key"], item["key_count"]) for item in results.results or []]


def _create_where_clause(data_sets: list[int] | None, hierarchies: list[int] | None) -> str:
    conditions = []
    if data_sets:
        if not all(isinstance(item, int) for item in data_sets):
            raise ToolkitValueError("All items in data_sets must be integers for SQL filtering.")
        conditions.append(f"dataSetId IN ({','.join(map(str, data_sets))})")
    if hierarchies:
        if not all(isinstance(item, int) for item in hierarchies):
            raise ToolkitValueError("All items in hierarchies must be integers for SQL filtering.")
        conditions.append(f"rootId IN ({','.join(map(str, hierarchies))})")

    if not conditions:
        return ""
    return f"\n         WHERE {' AND '.join(conditions)}"


def label_count(
    client: ToolkitClient,
    resource: Literal["assets", "events", "files", "timeseries", "sequences"],
    data_sets: list[int] | None = None,
    hierarchies: list[int] | None = None,
) -> list[tuple[str, int]]:
    """Get the label counts for a given resource.

    Args:
        client: ToolkitClient instance
        resource: The resource to get the label counts for. Can be one of "assets", "events", "files", "timeseries", or "sequences".
        data_sets: A list of data set IDs to filter by. If None, no filtering is applied.
        hierarchies: A list of hierarchy IDs to filter by. If None, no filtering is applied.

    Returns:
        A list of tuples with the label and its count.
    """
    if hierarchies and resource != "assets":
        raise ToolkitValueError(
            f"Hierarchies filtering for labels are only supported for assets, but {resource} was provided."
        )
    where_clause = _create_where_clause(data_sets, hierarchies)

    query = f"""WITH labels as (SELECT explode(labels) AS label
	FROM _cdf.{resource}{where_clause}
  )
SELECT label, COUNT(label) as label_count
FROM labels
GROUP BY label
ORDER BY label_count DESC;
"""
    results = client.transformations.preview(query, convert_to_string=False, limit=None, source_limit=None)
    # We know from the SQL that the result is a list of dictionaries with string keys and int values.
    return [(item["label"], item["label_count"]) for item in results.results or []]


@dataclass
class RelationshipCount:
    source_type: str
    target_type: str
    count: int

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        return cls(
            source_type=data["sourceType"],
            target_type=data["targetType"],
            count=int(data["relationshipCount"]),
        )


def relationship_aggregate_count(client: ToolkitClient, data_sets: list[int] | None = None) -> list[RelationshipCount]:
    """Get the relationship counts for the asset-centric resources.

    Args:
        client: ToolkitClient instance
        data_sets: A list of data set IDs to filter by. If None, no filtering is applied.

    """
    where_clause = ""
    if data_sets is not None:
        where_clause = f"\n         WHERE dataSetId IN ({','.join(map(str, data_sets))})"

    query = f"""SELECT
    sourceType,
    targetType,
    COUNT(externalId) AS relationshipCount
FROM
    _cdf.relationships{where_clause}
GROUP BY
    sourceType,
    targetType
"""
    results = client.transformations.preview(query, convert_to_string=False, limit=None, source_limit=None)
    return [
        RelationshipCount(item["sourceType"], item["targetType"], item["relationshipCount"])
        for item in results.results or []
    ]


def label_aggregate_count(client: ToolkitClient, data_sets: list[int] | None = None) -> int:
    """Get the total count of labels in the CDF project.

    Args:
        client: ToolkitClient instance
        data_sets: A list of data set IDs to filter by. If None, no filtering is applied.

    Returns:
        The total count of labels across all resources in the CDF project.
    """
    where_clause = ""
    if data_sets is not None:
        where_clause = f"\n         WHERE dataSetId IN ({','.join(map(str, data_sets))})"

    query = f"""SELECT
    COUNT(externalId) AS labelCount
FROM
    _cdf.labels{where_clause}"""

    results = client.transformations.preview(query, convert_to_string=False, limit=None, source_limit=None)
    if results.results:
        return int(results.results[0]["labelCount"])
    return 0


@dataclass
class ThrottlerState:
    _FILENAME = "tk-last-raw_count-{project}.txt"

    lock: BaseFileLock
    project: str
    _allowed_this_run: bool | None = None

    @classmethod
    def get(cls, project: str) -> "ThrottlerState":
        if project in _STATE_BY_PROJECT:
            return _STATE_BY_PROJECT[project]

        with _STATE_LOCK:
            # Re-check after acquiring the lock to handle the race condition.
            if project in _STATE_BY_PROJECT:
                return _STATE_BY_PROJECT[project]
            filepath = cls._filepath(project)
            # filelock needs a string path
            lock = FileLock(str(filepath.with_suffix(".lock")), timeout=10.0)
            state = cls(lock=lock, project=project)
            _STATE_BY_PROJECT[project] = state
            return state

    @classmethod
    def _filepath(cls, project: str) -> Path:
        # Enure the project name is a valid string for hashing.
        project = sanitize_filename(project)
        filename = cls._FILENAME.format(project=project)
        return Path(tempfile.gettempdir()) / filename

    def throttle(self) -> None:
        """Checks if the function can be run, and if so, updates the last call timestamp.

        This is an atomic operation protected by a file lock.

        Raises:
            ToolkitThrottledError: If the function has been called too recently.
        """
        try:
            with self.lock:
                now = time.time()
                filepath = self._filepath(self.project)
                if self._allowed_this_run is None:
                    last_call_epoch = 0.0
                    if filepath.exists():
                        try:
                            last_call_epoch = float(filepath.read_text(encoding="utf-8"))
                        except (ValueError, FileNotFoundError):
                            # File is corrupt or was deleted after check. Allow to run and overwrite.
                            pass

                    to_wait = (MAX_RUN_QUERY_FREQUENCY_MIN * 60) - (now - last_call_epoch)
                    if to_wait > 0:
                        raise ToolkitThrottledError(
                            f"Row count is limited to once every {MAX_RUN_QUERY_FREQUENCY_MIN} minutes. "
                            f"Please wait {to_wait:.2f} seconds before calling again.",
                            to_wait,
                        )
                    self._allowed_this_run = True

                # We are allowed to run, so we update the timestamp.
                filepath.write_text(str(now), encoding="utf-8")
        except Timeout:
            raise ToolkitError(
                "Could not acquire lock for throttling raw row count. "
                "Another toolkit process might be running. Please try again."
            )


_STATE_LOCK = RLock()
_STATE_BY_PROJECT: dict[str, ThrottlerState] = {}


def raw_row_count(client: ToolkitClient, raw_table_id: RawTable, max_count: int = MAX_ROW_ITERATION_RUN_QUERY) -> int:
    """Get the number of rows in a raw table.

    Args:
        client: ToolkitClient instance
        raw_table_id: The ID of the raw table to count rows in.
        max_count: The maximum number of rows to count. If the table has more rows than this, the count will be capped.

    Returns:
        The number of rows in the raw table.

    Raises:
        ToolkitThrottledError: If the row count is limited to once every MAX_RUN_QUERY_FREQUENCY_MIN minutes.
        ValueError: If max_count is not between 0 and MAX_ROW_ITERATION_RUN_QUERY (inclusive).
    """
    ThrottlerState.get(client.config.project).throttle()

    if not 0 <= max_count <= MAX_ROW_ITERATION_RUN_QUERY:
        raise ValueError(f"max_count must be between 0 and {MAX_ROW_ITERATION_RUN_QUERY} (inclusive).")

    query = f"""SELECT COUNT(key) AS row_count
FROM (
    SELECT key
    FROM `{raw_table_id.db_name}`.`{raw_table_id.table_name}`
    LIMIT {max_count}
) AS limited_keys"""

    results = client.transformations.preview(query, convert_to_string=False, limit=None, source_limit=None)
    if results.results:
        return int(results.results[0]["row_count"])
    return 0



import re
from typing import Literal

from pydantic import Field, field_validator

from cognite_toolkit._cdf_tk.utils import humanize_collection

from .base import ToolkitResource
from .view_field_definitions import ViewReference

JSON_PATH_PATTERN = re.compile(
    r"^([a-zA-Z_][a-zA-Z0-9_]*|\[\d+\])(\.[a-zA-Z_][a-zA-Z0-9_]*|\[\d+\]|\[\'[^\']*\'\]|\[\"[^\"]*\"\])*$"
)


class ResourceViewMappingYAML(ToolkitResource):
    resource_type: Literal["asset", "event", "file", "timeSeries", "sequence", "assetAnnotation", "fileAnnotation"] = (
        Field(
            description="The type of the resource to map to the view.",
        )
    )
    view_id: ViewReference = Field(
        description="The view to map the resource to.",
    )
    property_mapping: dict[str, str] = Field(
        description="A dictionary mapping from resource property to view property.",
    )

    @field_validator("property_mapping")
    @classmethod
    def validate_json_paths(cls, value: dict[str, str]) -> dict[str, str]:
        if not isinstance(value, dict):
            return value
        not_matching_keys = [k for k in value.keys() if not re.match(JSON_PATH_PATTERN, k)]
        if not_matching_keys:
            raise ValueError(f"Invalid JSON paths: {humanize_collection(not_matching_keys)}")
        return value
